# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import copy
from typing import Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np

from orix.crystal_map.crystal_map_properties import CrystalMapProperties
from orix.crystal_map.phase_list import ALL_COLORS, Phase, PhaseList
from orix.quaternion import Orientation, Rotation


class CrystalMap:
    """Crystallographic map of orientations, crystal phases and key
    properties associated with every spatial coordinate in a 1D or 2D.

    Parameters
    ----------
    rotations
        Rotation in each point. Must be passed with all spatial
        dimensions in the first array axis (flattened). May contain
        multiple rotations per point, included in the second array
        axes. Crystal map data size is set equal to the first array
        axis' size.
    phase_id
        Phase ID of each pixel. IDs equal to ``-1`` are considered not
        indexed. If not given, all points are considered to belong to
        one phase with ID ``0``.
    x
        Map x coordinate of each data point. If not given, the map is
        assumed to be 1D, and it is set to an array of increasing
        integers from 0 to the length of the ``phase_id`` array.
    y
        Map y coordinate of each data point. If not given, the map is
        assumed to be 1D, and it is set to ``None``.
    phase_list
        A list of phases in the data with their with names, space
        groups, point groups, and structures. The order in which the
        phases appear in the list is important, as it is this, and not
        the phases' IDs, that is used to link the phases to the
        input ``phase_id`` if the IDs aren't exactly the same as in
        ``phase_id``. If not given, a phase list with as many phases as
        there are unique phase IDs in ``phase_id`` is created.
    prop
        Dictionary of properties of each data point.
    scan_unit
        Length unit of the data. Default is ``"px"``.
    is_in_data
        Array of booleans signifying whether a point is in the data.

    See Also
    --------
    create_coordinate_arrays
    :mod:`~orix.data`
    :func:`~orix.io.load`

    Notes
    -----
    Data is stored as 1D arrays and reshaped when necessary.

    Examples
    --------
    Constructing a crystal map from scratch, with two rows and three
    columns and containing Austenite and Ferrite orientations

    >>> from diffpy.structure import Atom, Lattice, Structure
    >>> from orix.crystal_map import create_coordinate_arrays, CrystalMap, PhaseList
    >>> from orix.quaternion import Rotation
    >>> coords, n = create_coordinate_arrays(shape=(2, 3))
    >>> structures = [
    ...     Structure(
    ...         title="austenite",
    ...         atoms=[Atom("fe", [0] * 3)],
    ...         lattice=Lattice(0.360, 0.360, 0.360, 90, 90, 90)
    ...     ),
    ...     Structure(
    ...         title="ferrite",
    ...         atoms=[Atom("fe", [0] * 3)],
    ...         lattice=Lattice(0.287, 0.287, 0.287, 90, 90, 90)
    ...     )
    ... ]
    >>> xmap = CrystalMap(
    ...     rotations=Rotation.from_axes_angles([0, 0, 1], np.linspace(0, np.pi, n)),
    ...     phase_id=np.array([0, 0, 1, 1, 0, 1]),
    ...     x=coords["x"],
    ...     y=coords["y"],
    ...     phase_list=PhaseList(space_groups=[225, 229], structures=structures),
    ...     prop={"score": np.random.random(n)},
    ...     scan_unit="nm",
    ... )
    >>> xmap
    Phase  Orientations       Name  Space group  Point group  Proper point group       Color
        0     3 (50.0%)  austenite        Fm-3m         m-3m                 432    tab:blue
        1     3 (50.0%)    ferrite        Im-3m         m-3m                 432  tab:orange
    Properties: score
    Scan unit: nm

    Data in a crystal map can be selected in multiple ways. Let's
    demonstrate this on a dual phase dataset available in the
    :mod:`~orix.data` module

    >>> from orix import data
    >>> xmap = data.sdss_ferrite_austenite(allow_download=True)
    >>> xmap
    Phase   Orientations       Name  Space group  Point group  Proper point group       Color
        1   5657 (48.4%)  austenite         None          432                 432    tab:blue
        2   6043 (51.6%)    ferrite         None          432                 432  tab:orange
    Properties: iq, dp
    Scan unit: um
    >>> xmap.shape
    (100, 117)

    Selecting based on coordinates, passing ranges (slices), integers or
    both

    >>> xmap2 = xmap[20:40, 50:60]
    >>> xmap2
    Phase  Orientations       Name  Space group  Point group  Proper point group       Color
        1   148 (74.0%)  austenite         None          432                 432    tab:blue
        2    52 (26.0%)    ferrite         None          432                 432  tab:orange
    Properties: iq, dp
    Scan unit: um
    >>> xmap2.shape
    (20, 10)
    >>> xmap2 = xmap[20:40, 3]
    >>> xmap2
    Phase  Orientations       Name  Space group  Point group  Proper point group       Color
        1    16 (80.0%)  austenite         None          432                 432    tab:blue
        2     4 (20.0%)    ferrite         None          432                 432  tab:orange
    Properties: iq, dp
    Scan unit: um
    >>> xmap2.shape
    (20, 1)

    Note that 1-dimensions are NOT removed

    >>> xmap2 = xmap[10, 10]
    >>> xmap2
    Phase  Orientations     Name  Space group  Point group  Proper point group       Color
        2    1 (100.0%)  ferrite         None          432                 432  tab:orange
    Properties: iq, dp
    Scan unit: um
    >>> xmap2.shape
    (1, 1)

    Select by phase name(s)

    >>> xmap2 = xmap["austenite"]
    >>> xmap2
    Phase  Orientations       Name  Space group  Point group  Proper point group     Color
        1  5657 (100.0%)  austenite         None          432                 432  tab:blue
    Properties: iq, dp
    Scan unit: um
    >>> xmap2.shape
    (100, 117)
    >>> xmap["austenite", "ferrite"]
    Phase  Orientations       Name  Space group  Point group  Proper point group       Color
        1  5657 (48.4%)  austenite         None          432                 432    tab:blue
        2  6043 (51.6%)    ferrite         None          432                 432  tab:orange
    Properties: iq, dp
    Scan unit: um

    Select by indexed and not indexed data

    >>> xmap["indexed"]
    Phase  Orientations       Name  Space group  Point group  Proper point group       Color
        1  5657 (48.4%)  austenite         None          432                 432    tab:blue
        2  6043 (51.6%)    ferrite         None          432                 432  tab:orange
    Properties: iq, dp
    Scan unit: um
    >>> xmap["not_indexed"]
    No data.

    Select with a boolean array (possibly chained)

    >>> xmap[xmap.dp > 0.81]
    Phase  Orientations       Name  Space group  Point group  Proper point group       Color
        1  4092 (44.8%)  austenite         None          432                 432    tab:blue
        2  5035 (55.2%)    ferrite         None          432                 432  tab:orange
    Properties: iq, dp
    Scan unit: um
    >>> xmap[(xmap.iq > np.mean(xmap.iq)) & (xmap.phase_id == 1)]
    Phase  Orientations       Name  Space group  Point group  Proper point group     Color
        1  1890 (100.0%)  austenite         None          432                 432  tab:blue
    Properties: iq, dp
    Scan unit: um
    """

    def __init__(
        self,
        rotations: Rotation,
        phase_id: Optional[np.ndarray] = None,
        x: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        phase_list: Optional[PhaseList] = None,
        prop: Optional[dict] = None,
        scan_unit: Optional[str] = "px",
        is_in_data: Optional[np.ndarray] = None,
    ):
        # Set rotations
        if not isinstance(rotations, Rotation):
            raise ValueError(
                f"rotations must be of type {Rotation}, not {type(rotations)}."
            )
        self._rotations = rotations

        # Set data size
        data_size = rotations.shape[0]

        # Set phase IDs
        if phase_id is None:  # Assume single phase data
            phase_id = np.zeros(data_size)
        phase_id = phase_id.astype(int)
        self._phase_id = phase_id

        # Set data point IDs
        point_id = np.arange(data_size)
        self._id = point_id

        # Set spatial coordinates
        if x is None and y is None:
            x = np.arange(data_size)
        self._x = x
        self._y = y

        # Create phase list
        # Sorted in ascending order
        unique_phase_ids = np.unique(phase_id)
        include_not_indexed = False
        if unique_phase_ids[0] == -1:
            include_not_indexed = True
            unique_phase_ids = unique_phase_ids[1:]
        # Also sorted in ascending order
        if phase_list is None:
            self._phases = PhaseList(ids=unique_phase_ids)
        else:
            phase_list = phase_list.deepcopy()
            phase_ids = phase_list.ids
            n_different = len(phase_ids) - len(unique_phase_ids)
            if n_different > 0:
                # Remove superfluous phases by removing the phases whose
                # ID is not in the ID array, in descending list order
                for i in phase_ids[::-1]:
                    if i not in unique_phase_ids:
                        del phase_list[i]
                        n_different -= 1
                    if n_different == 0:
                        break
            elif n_different < 0:
                # Create new phase list adding the missing phases with
                # default initial values (but unique colors)
                phase_dict = {}
                all_colors = list(ALL_COLORS.keys())
                all_unique_colors = np.delete(
                    all_colors, np.isin(all_colors, phase_list.colors)
                )
                ci = 0
                for i in unique_phase_ids:
                    if i in phase_ids:
                        phase_dict[i] = phase_list[i]
                    else:
                        phase_dict[i] = Phase(color=all_unique_colors[ci])
                        ci += 1
                phase_list = PhaseList(phase_dict)
            # Ensure phase list IDs correspond to IDs in phase_id array
            new_ids = list(unique_phase_ids.astype(int))
            phase_list._dict = dict(zip(new_ids, phase_list._dict.values()))
            self._phases = phase_list

        # Set whether measurements are indexed
        is_indexed = np.ones(data_size, dtype=bool)
        is_indexed[np.where(phase_id == -1)] = False

        # Add "not_indexed" to phase list and ensure not indexed points
        # have correct phase ID
        if include_not_indexed:
            self.phases.add_not_indexed()
            self._phase_id[~is_indexed] = -1

        # Set array with True for points in data
        if is_in_data is None:
            is_in_data = np.ones(data_size, dtype=bool)
        self.is_in_data = is_in_data

        # Set scan unit
        self.scan_unit = scan_unit

        # Set properties
        if prop is None:
            prop = {}
        self._prop = CrystalMapProperties(prop, id=point_id)

        # Set original data shape (needed if data shape changes in
        # __getitem__())
        self._original_shape = self._data_shape_from_coordinates(only_is_in_data=False)

    @property
    def id(self) -> np.ndarray:
        """Return the ID of points in data."""
        return self._id[self.is_in_data]

    @property
    def size(self) -> int:
        """Return the total number of points in data."""
        return np.count_nonzero(self.is_in_data)

    @property
    def shape(self) -> tuple:
        """Return the shape of points in data."""
        return self._data_shape_from_coordinates()

    @property
    def ndim(self) -> int:
        """Return the number of data dimensions of points in data."""
        return len(self.shape)

    @property
    def x(self) -> Union[None, np.ndarray]:
        """Return the x coordinates of points in data."""
        if self._x is None or len(np.unique(self._x)) == 1:
            return
        else:
            return self._x[self.is_in_data]

    @property
    def y(self) -> Union[None, np.ndarray]:
        """Return the y coordinates of points in data."""
        if self._y is None or len(np.unique(self._y)) == 1:
            return
        else:
            return self._y[self.is_in_data]

    @property
    def dx(self) -> float:
        """Return the x coordinate step size."""
        return self._step_size_from_coordinates(self._x)

    @property
    def dy(self) -> float:
        """Return the y coordinate step size."""
        return self._step_size_from_coordinates(self._y)

    @property
    def row(self) -> Union[None, np.ndarray]:
        """Return the row coordinate of each point in the data.

        Returns ``None`` if :attr:`z` is not ``None``.

        Examples
        --------
        >>> from orix.crystal_map import CrystalMap
        >>> xmap = CrystalMap.empty((3, 4))
        >>> xmap.row
        array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])
        >>> xmap[1:3, 1:3].row
        array([0, 0, 1, 1])
        """
        orig_shape = self._original_shape
        if len(orig_shape) == 1:
            if self.x is None:
                orig_shape += (1,)
            else:
                orig_shape = (1,) + orig_shape
        rows, _ = np.indices(orig_shape)
        rows = rows.flatten()[self.is_in_data]
        rows -= rows.min()
        return rows

    @property
    def col(self) -> Union[None, np.ndarray]:
        """Return the column coordinate of each point in the data.

        Returns ``None`` if :attr:`z` is not ``None``.

        Examples
        --------
        >>> from orix.crystal_map import CrystalMap
        >>> xmap = CrystalMap.empty((3, 4))
        >>> xmap.col
        array([0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3])
        >>> xmap[1:3, 1:3].col
        array([0, 1, 0, 1])
        """
        shape = self._original_shape
        if len(shape) == 1:
            if self.x is None:
                shape += (1,)
            else:
                shape = (1,) + shape
        _, cols = np.indices(shape)
        cols = cols.flatten()[self.is_in_data]
        cols -= cols.min()
        return cols

    @property
    def phase_id(self) -> np.ndarray:
        """Return or set the phase IDs of points in data.

        Parameters
        ----------
        value : numpy.ndarray or int
            Phase ID of points in data.
        """
        return self._phase_id[self.is_in_data]

    @phase_id.setter
    def phase_id(self, value: Union[np.ndarray, int]):
        """Set phase ID of points in data."""
        self._phase_id[self.is_in_data] = value
        if value == -1 and "not_indexed" not in self.phases.names:
            self.phases.add_not_indexed()

    @property
    def phases(self) -> PhaseList:
        """Return or set the list of phases.

        Parameters
        ----------
        value : PhaseList
            Phase list with at least as many phases as unique phase IDs
            in :attr:`phase_id`.

        Raises
        ------
        ValueError
            If there are fewer phases in the list than unique phase IDs.
        """
        return self._phases

    @phases.setter
    def phases(self, value: PhaseList):
        """Set the list of phases."""
        if np.unique(self.phase_id).size > value.size:
            raise ValueError(
                "There must be at least as many phases as there are unique phase IDs"
            )
        else:
            self._phases = value

    @property
    def phases_in_data(self) -> PhaseList:
        """Return the list of phases in data.

        See Also
        --------
        phases

        Notes
        -----
        Can be useful when there are phases in :attr:`phases` which are
        not in the data.
        """
        unique_ids = np.unique(self.phase_id)
        phase_list = self.phases[np.intersect1d(unique_ids, self.phases.ids)]
        if isinstance(phase_list, Phase):  # One phase in data
            # Get phase ID so it carries over to the new `PhaseList`
            # instance
            phase = phase_list  # Since it's actually a single phase
            phase_id = self.phases.id_from_name(phase.name)
            return PhaseList(phases=phase, ids=phase_id)
        else:  # Multiple phases in data
            return phase_list

    @property
    def rotations(self) -> Rotation:
        """Return the rotations in the data."""
        return self._rotations[self.is_in_data]

    @property
    def rotations_per_point(self) -> int:
        """Return the number of rotations per data point in data."""
        return self.rotations.size // self.is_indexed.size

    @property
    def rotations_shape(self) -> tuple:
        """Return the shape of :attr:`rotations`.

        Notes
        -----
        Map shape and possible multiple rotations per point are
        accounted for. 1-dimensions are squeezed out.
        """
        return tuple(i for i in self.shape + (self.rotations_per_point,) if i != 1)

    @property
    def orientations(self) -> Orientation:
        """Return orientations (rotations respecting symmetry), in data.

        Raises
        ------
        ValueError
            When the (potentially sliced map) has more than one phase in
            the data.

        Examples
        --------
        >>> from orix import data
        >>> xmap = data.sdss_ferrite_austenite(allow_download=True)
        >>> xmap
        Phase   Orientations       Name  Space group  Point group  Proper point group       Color
            1   5657 (48.4%)  austenite         None          432                 432    tab:blue
            2   6043 (51.6%)    ferrite         None          432                 432  tab:orange
        Properties: iq, dp
        Scan unit: um
        >>> xmap["austenite"].orientations
        Orientation (5657,) 432
        [[ 0.8686  0.3569 -0.2749 -0.2064]
         [ 0.8681  0.3581 -0.2744 -0.2068]
         [ 0.8684  0.3578 -0.2751 -0.2052]
         ...
         [ 0.9639  0.022   0.0754 -0.2545]
         [ 0.8854  0.3337 -0.2385  0.2187]
         [ 0.885   0.3341 -0.2391  0.2193]]
        """
        phases = self.phases_in_data
        if phases.size == 1:
            # Extract top matching rotations per point, if more than one
            if self.rotations_per_point > 1:
                rotations = self.rotations[:, 0]
            else:
                rotations = self.rotations
            # Point group can be None, so it cannot be passed upon
            # initialization to Orientation but has to be set afterwards
            # to trigger the checks
            orientations = Orientation(rotations)
            orientations.symmetry = phases[:].point_group
            return orientations
        else:
            raise ValueError(
                f"Data has the phases {phases.names}, however, you are executing a "
                "command that only permits one phase."
            )

    @property
    def is_indexed(self) -> np.ndarray:
        """Return whether points in data are indexed."""
        return self.phase_id != -1

    @property
    def all_indexed(self) -> np.ndarray:
        """Return whether all points in data are indexed."""
        return np.count_nonzero(self.is_indexed) == self.is_indexed.size

    @property
    def prop(self) -> CrystalMapProperties:
        """Return the data properties in each data point."""
        self._prop.is_in_data = self.is_in_data
        self._prop.id = self.id
        return self._prop

    @property
    def _coordinates(self) -> dict:
        """Return the coordinates of points in the data."""
        return {"y": self.y, "x": self.x}

    @property
    def _all_coordinates(self) -> dict:
        """Return the coordinates of all points."""
        return {"y": self._y, "x": self._x}

    @property
    def _step_sizes(self) -> dict:
        """Return the step sizes of dimensions in the data."""
        return {"y": self.dy, "x": self.dx}

    @property
    def _coordinate_axes(self) -> dict:
        """Return which data axis corresponds to which coordinate."""
        present_coordinates = [k for k, v in self._coordinates.items() if v is not None]
        return {i: coord for i, coord in zip(range(self.ndim), present_coordinates)}

    def __getattr__(self, item):
        """Return an attribute in the :attr:`prop` dictionary directly
        from the ``CrystalMap`` instance.

        Called when the default attribute access fails with an
        ``AttributeError``.
        """
        if item in self.__getattribute__("_prop"):
            # Calls CrystalMapProperties.__getitem__()
            return self.prop[item]
        else:
            return object.__getattribute__(self, item)

    def __setattr__(self, name, value):
        """Set a class instance attribute."""
        if hasattr(self, "_prop") and name in self._prop:
            # Calls CrystalMapProperties.__setitem__()
            self.prop[name] = value
        else:
            return object.__setattr__(self, name, value)

    def __getitem__(self, key: Union[str, slice, tuple, int, np.ndarray]):
        """Get a masked copy of the CrystalMap instance.

        See the docstring of ``__init__()`` for examples.

        Parameters
        ----------
        key
            If ``str``, it must be a valid phase or ``"not_indexed"`` or
            ``"indexed"``. If ``slice`` or ``tuple``, it must be within
            the map shape. If ``int``, it must be a valid
            :attr:`self.id`. If boolean array, it must be of map shape.
        """
        # Initiate a mask to be added to the returned copy of the
        # CrystalMap instance, to ensure that only the unmasked values
        # are in the data of the copy (True in `is_in_data`). First, no
        # points are in the data, but are added if they satisfy the
        # condition in the input key.
        is_in_data = np.zeros(self.size, dtype=bool)

        # The original mask might already have set some points to not be
        # in the data. If so, `is_in_data` is used to update the
        # original `is_in_data`. Since `new_is_in_data` is not initiated
        # for all key types, we declare it here and check for it later.
        new_is_in_data = None

        # Override mask values
        if isinstance(key, str) or (isinstance(key, tuple) and isinstance(key[0], str)):
            # From phase string(s)
            if not isinstance(key, tuple):  # Make single string iterable
                key = (key,)
            for k in key:
                for phase_id, phase in self.phases:
                    if k == phase.name:
                        is_in_data[self.phase_id == phase_id] = True
                    elif k.lower() == "indexed":
                        # Add all indexed phases to data
                        is_in_data[self.phase_id != -1] = True
        elif isinstance(key, np.ndarray) and key.dtype == np.bool_:
            # From boolean numpy array
            is_in_data = key
        elif isinstance(key, (slice, int)) or (
            isinstance(key, tuple)
            and any([(isinstance(i, slice) or isinstance(i, int)) for i in key])
        ):
            # From slice(s) or int
            if isinstance(key, (slice, int)):
                key = (key,)

            slices = [slice(None, None, None)] * self.ndim
            for i, k in enumerate(key):
                slices[i] = k

            new_is_in_data_slice = np.zeros(self.shape, dtype=bool)  # > 1D
            new_is_in_data_slice[tuple(slices)] = True

            # Insert new (sub)mask into old full mask
            new_is_in_data = self.is_in_data.reshape(self._original_shape).copy()
            new_is_in_data[self._data_slices_from_coordinates()] = new_is_in_data_slice
            new_is_in_data = new_is_in_data.ravel()

        # Insert the mask into a mask with the full map shape, if not
        # done already
        if new_is_in_data is None:
            new_is_in_data = np.zeros_like(self.is_in_data, dtype=bool)  # 1D
            new_is_in_data[self.id] = is_in_data

        # Return a copy with all attributes shallow except for the mask
        new_map = copy.copy(self)
        new_map.is_in_data = new_is_in_data

        return new_map

    def __repr__(self) -> str:
        """Return a nice representation of the data."""
        if self.size == 0:
            return "No data."

        phases = self.phases_in_data
        phase_ids = self.phase_id

        # Ensure attributes set to None are treated OK
        names = ["None" if not name else name for name in phases.names]
        sg_names = ["None" if not i else i.short_name for i in phases.space_groups]
        pg_names = ["None" if not i else i.name for i in phases.point_groups]
        ppg_names = [
            "None" if not i else i.proper_subgroup.name for i in phases.point_groups
        ]

        # Determine column widths
        unique_phases = np.unique(phase_ids)
        p_sizes = [np.where(phase_ids == i)[0].size for i in unique_phases]
        id_len = 5
        ori_len = max(max([len(str(p_size)) for p_size in p_sizes]) + 9, 12)
        name_len = max(max([len(n) for n in names]), 4)
        sg_len = max(max([len(i) for i in sg_names]), 11)
        pg_len = max(max([len(i) for i in pg_names]), 11)
        ppg_len = max(max([len(i) for i in ppg_names]), 18)
        col_len = max(max([len(i) for i in phases.colors]), 5)

        # Column alignment
        align = ">"  # right ">" or left "<"

        # Header (note the two-space spacing)
        representation = (
            "{:{align}{width}}  ".format("Phase", width=id_len, align=align)
            + "{:{align}{width}}  ".format("Orientations", width=ori_len, align=align)
            + "{:{align}{width}}  ".format("Name", width=name_len, align=align)
            + "{:{align}{width}}  ".format("Space group", width=sg_len, align=align)
            + "{:{align}{width}}  ".format("Point group", width=pg_len, align=align)
            + "{:{align}{width}}  ".format(
                "Proper point group", width=ppg_len, align=align
            )
            + "{:{align}{width}}\n".format("Color", width=col_len, align=align)
        )

        # Overview of data for each phase
        for i, phase_id in enumerate(unique_phases.astype(int)):
            p_size = np.where(phase_ids == phase_id)[0].size
            p_fraction = 100 * p_size / self.size
            ori_str = f"{p_size} ({p_fraction:.1f}%)"
            representation += (
                f"{phase_id:{align}{id_len}}  "
                + f"{ori_str:{align}{ori_len}}  "
                + f"{names[i]:{align}{name_len}}  "
                + f"{sg_names[i]:{align}{sg_len}}  "
                + f"{pg_names[i]:{align}{pg_len}}  "
                + f"{ppg_names[i]:{align}{ppg_len}}  "
                + f"{phases.colors[i]:{align}{col_len}}\n"
            )

        # Properties and spatial coordinates
        props = []
        for k in self.prop.keys():
            props.append(k)
        representation += "Properties: " + ", ".join(props) + "\n"

        # Scan unit
        representation += f"Scan unit: {self.scan_unit}"

        return representation

    def deepcopy(self) -> "CrystalMap":
        """Return a deep copy using :func:`copy.deepcopy` function."""
        return copy.deepcopy(self)

    @classmethod
    def empty(
        cls,
        shape: Union[None, int, tuple] = None,
        step_sizes: Union[None, int, tuple] = None,
    ) -> "CrystalMap":
        """Return a crystal map of a given 2D shape and step sizes with
        identity rotations.

        Parameters
        ----------
        shape
            Map shape. Default is a 2D map of shape (5, 10), i.e. with
            five rows and ten columns.
        step_sizes
            Map step sizes. If not given, it is set to 1 px in each map
            direction given by ``shape``.

        Returns
        -------
        xmap
            Crystal map.
        """
        d, n = create_coordinate_arrays(shape, step_sizes)
        d["rotations"] = Rotation.identity((n,))
        return cls(**d)

    def get_map_data(
        self,
        item: Union[str, np.ndarray],
        decimals: Optional[int] = None,
        fill_value: Union[int, float, None] = np.nan,
    ) -> np.ndarray:
        """Return an array of a class instance attribute, with values
        equal to ``False`` in :attr:`self.is_in_data` set to
        ``fill_value``, of map data shape.

        Parameters
        ----------
        item
            Name of the class instance attribute or a
            :class:`numpy.ndarray`.
        decimals
            Number of decimals to round data point values to. If not
            given, no rounding is done.
        fill_value
            Value to fill points not in the data with. Default is
            :class:`numpy.nan`.

        Returns
        -------
        output_array
            Array of the class instance attribute with points not in
            data set to ``fill_value``, of float data type.

        Notes
        -----
        Rotations and orientations should be accessed via
        :attr:`rotations` and :attr:`orientations`.

        If ``item`` is ``"orientations"`` or ``"rotations"`` and there
        are multiple rotations per point, only the first rotation is
        used. Rotations are returned as Euler angles.

        Examples
        --------
        >>> from orix import data
        >>> xmap = data.sdss_ferrite_austenite(allow_download=True)
        >>> xmap
        Phase   Orientations       Name  Space group  Point group  Proper point group       Color
            1   5657 (48.4%)  austenite         None          432                 432    tab:blue
            2   6043 (51.6%)    ferrite         None          432                 432  tab:orange
        Properties: iq, dp
        Scan unit: um
        >>> xmap.shape
        (100, 117)

        Get a 2D map in the correct shape of any attribute, ready for
        plotting

        >>> xmap.iq.shape
        (11700,)
        >>> iq = xmap.get_map_data("iq")
        >>> iq.shape
        (100, 117)
        """
        # Get full map shape
        map_shape = self._original_shape

        # Declare array of correct shape, accounting for RGB
        # TODO: Better account for `item.shape`, e.g. quaternions
        #  (item.shape[-1] == 4) in a more general way than here (not
        #  more if/else)!
        map_size = np.prod(map_shape)
        if isinstance(item, np.ndarray):
            array = np.empty(map_size, dtype=item.dtype)
            if item.shape[-1] == 3 and map_size > 3:  # Assume RGB
                map_shape += (3,)
                array = np.column_stack((array,) * 3)
        elif item in ["orientations", "rotations"]:  # Definitely RGB
            array = np.empty(map_size, dtype=np.float64)
            map_shape += (3,)
            array = np.column_stack((array,) * 3)
        else:
            array = np.empty(map_size, dtype=np.float64)

        # Enter non-masked values into array
        if isinstance(item, np.ndarray):
            # TODO: Account for 2D map with more than one value per point
            array[self.is_in_data] = item
        elif item in ["orientations", "rotations"]:
            if item == "rotations":
                # Use only the top matching rotation per point
                if self.rotations_per_point > 1:
                    rotations = self.rotations[:, 0]
                else:
                    rotations = self.rotations
                array[self.is_in_data] = rotations.to_euler()
            else:  # item == "orientations"
                # Fill in orientations per phase
                for i, _ in self.phases_in_data:
                    phase_mask = (self._phase_id == i) * self.is_in_data
                    phase_mask_in_data = self.phase_id == i
                    array[phase_mask] = self[phase_mask_in_data].orientations.to_euler()
        else:  # String
            data = self.__getattr__(item)
            if data is None:
                raise ValueError(f"{item} is {data}.")
            else:
                # TODO: Account for 2D map with more than one value per point
                array[self.is_in_data] = data
                array = array.astype(data.dtype)

        # Slice and reshape array
        slices = self._data_slices_from_coordinates()
        reshaped_array = array.reshape(map_shape)
        sliced_array = reshaped_array[slices]

        # Reshape and slice mask with points not in data
        if array.shape[-1] == 3 and map_size > 3:  # RGB
            not_in_data = np.dstack((~self.is_in_data,) * 3)
        else:  # Scalar
            not_in_data = ~self.is_in_data
        not_in_data = not_in_data.reshape(map_shape)[slices]

        # Fill points not in data with the fill value
        if not_in_data.any():
            if fill_value is None or fill_value is np.nan:
                sliced_array = sliced_array.astype(np.float64)
            sliced_array[not_in_data] = fill_value

        # Round values
        if decimals is not None:
            output_array = np.round(sliced_array, decimals=decimals)
        else:  # np.issubdtype(array.dtype, np.bool_):
            output_array = sliced_array

        return output_array

    def plot(
        self,
        value: Union[np.ndarray, str, None] = None,
        overlay: Union[str, np.ndarray, None] = None,
        scalebar: bool = True,
        scalebar_properties: Optional[dict] = None,
        legend: bool = True,
        legend_properties: Optional[dict] = None,
        colorbar: bool = False,
        colorbar_label: Optional[str] = None,
        colorbar_properties: Optional[dict] = None,
        remove_padding: bool = False,
        return_figure: bool = False,
        figure_kwargs: Optional[dict] = None,
        **kwargs,
    ) -> plt.Figure:
        r"""Plot a 2D map with any crystallographic map property as map
        values.

        Wraps :meth:`matplotlib.axes.Axes.imshow`: see that method for
        relevant keyword arguments.

        Parameters
        ----------
        value
            An array or an attribute string to plot. If not given, a
            phase map is plotted.
        overlay
            Name of map property or a property array to use in the
            alpha (RGBA) channel. The property range is adjusted for
            maximum contrast. Not used if not given.
        scalebar
            Whether to add a scalebar (default is ``True``) along the
            horizontal map dimension.
        scalebar_properties
            Keyword arguments passed to
            :class:`matplotlib_scalebar.scalebar.ScaleBar`.
        legend
            Whether to add a legend to the plot. This is only
            implemented for a phase plot (in which case default is
            ``True``).
        legend_properties
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.legend`.
        colorbar
            Whether to add an opinionated colorbar (default is
            ``False``).
        colorbar_label
            Label of colorbar.
        colorbar_properties
            Keyword arguments passed to
            :meth:`orix.plot.CrystalMapPlot.add_colorbar`.
        remove_padding
            Whether to remove white padding around figure (default is
            ``False``).
        return_figure
            Whether to return the figure (default is ``False``).
        figure_kwargs
            Keyword arguments passed to
            :func:`matplotlib.pyplot.subplots`.
        **kwargs
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.imshow`.

        Returns
        -------
        fig
            The created figure, returned if ``return_figure=True``.

        See Also
        --------
        matplotlib.axes.Axes.imshow
        orix.plot.CrystalMapPlot.plot_map
        orix.plot.CrystalMapPlot.add_scalebar
        orix.plot.CrystalMapPlot.add_overlay
        orix.plot.CrystalMapPlot.add_colorbar

        Examples
        --------
        >>> from orix import data
        >>> xmap = data.sdss_ferrite_austenite(allow_download=True)
        >>> xmap
        Phase   Orientations       Name  Space group  Point group  Proper point group       Color
            1   5657 (48.4%)  austenite         None          432                 432    tab:blue
            2   6043 (51.6%)    ferrite         None          432                 432  tab:orange
        Properties: iq, dp
        Scan unit: um

        Plot phase map

        >>> xmap.plot()

        Remove padding and return the figure (e.g. to be saved)

        >>> fig = xmap.plot(remove_padding=True, return_figure=True)

        Plot a dot product (similarity score) map

        >>> xmap.plot("dp", colorbar=True, colorbar_label="Dot product", cmap="gray")
        """
        # Register "plot_map" projection with Matplotlib
        import orix.plot.crystal_map_plot

        if figure_kwargs is None:
            figure_kwargs = dict()

        fig, ax = plt.subplots(subplot_kw=dict(projection="plot_map"), **figure_kwargs)
        ax.plot_map(
            self,
            value=value,
            scalebar=scalebar,
            scalebar_properties=scalebar_properties,
            legend=legend,
            legend_properties=legend_properties,
            **kwargs,
        )
        if overlay is not None:
            ax.add_overlay(self, overlay)
        if remove_padding:
            ax.remove_padding()
        if colorbar:
            if colorbar_properties is None:
                colorbar_properties = dict()
            ax.add_colorbar(label=colorbar_label, **colorbar_properties)
        if return_figure:
            return fig

    @staticmethod
    def _step_size_from_coordinates(coordinates: np.ndarray) -> float:
        """Return step size in input ``coordinates`` array.

        Parameters
        ----------
        coordinates
            Linear coordinate array.

        Returns
        -------
        step_size
            Step size in ``coordinates`` array.
        """
        unique_sorted = np.sort(np.unique(coordinates))
        step_size = 0
        if unique_sorted.size != 1:
            step_size = unique_sorted[1] - unique_sorted[0]
        return step_size

    def _data_slices_from_coordinates(self, only_is_in_data: bool = True) -> tuple:
        """Return a tuple of slices defining the current data extent in
        all directions.

        Parameters
        ----------
        only_is_in_data
            Whether to determine slices of points in data or all points.
            Default is ``True``.

        Returns
        -------
        slices
            Data slice in each existing dimension, in (z, y, x) order.
        """
        if only_is_in_data:
            coordinates = self._coordinates
        else:
            coordinates = self._all_coordinates

        # Loop over dimension coordinates and step sizes
        slices = []
        for coords, step in zip(coordinates.values(), self._step_sizes.values()):
            if coords is not None and step != 0:
                c_min, c_max = np.min(coords), np.max(coords)
                i_min = int(np.around(c_min / step))
                i_max = int(np.around((c_max / step) + 1))
                slices.append(slice(i_min, i_max))

        return tuple(slices)

    def _data_shape_from_coordinates(self, only_is_in_data: bool = True) -> tuple:
        """Return data shape based upon coordinate arrays.

        Parameters
        ----------
        only_is_in_data
            Whether to determine shape of points in data or all points.
            Default is ``True``.

        Returns
        -------
        data_shape
            Shape of data in all existing dimensions, in (z, y, x) order.
        """
        data_shape = []
        for dim_slice in self._data_slices_from_coordinates(only_is_in_data):
            data_shape.append(dim_slice.stop - dim_slice.start)
        return tuple(data_shape)


def create_coordinate_arrays(
    shape: Optional[tuple] = None, step_sizes: Optional[tuple] = None
) -> Tuple[dict, int]:
    """Create flattened coordinate arrays from a given map shape and
    step sizes, suitable for initializing a
    :class:`~orix.crystal_map.CrystalMap`. Arrays for 1D or 2D maps can
    be returned.

    Parameters
    ----------
    shape
        Map shape. Default is a 2D map of shape (5, 10) with five rows
        and ten columns.
    step_sizes
        Map step sizes. If not given, it is set to 1 px in each map
        direction given by ``shape``.

    Returns
    -------
    d
        Dictionary with keys ``"y"`` and ``"x"``, depending on the
        length of ``shape``, with coordinate arrays.
    map_size
        Number of map points.

    Examples
    --------
    >>> from orix.crystal_map import create_coordinate_arrays
    >>> create_coordinate_arrays((2, 3))
    ({'x': array([0, 1, 2, 0, 1, 2]), 'y': array([0, 0, 0, 1, 1, 1])}, 6)
    >>> create_coordinate_arrays((3, 2))
    ({'x': array([0, 1, 0, 1, 0, 1]), 'y': array([0, 0, 1, 1, 2, 2])}, 6)
    >>> create_coordinate_arrays((2, 3), (1.5, 1.5))
    ({'x': array([0. , 1.5, 3. , 0. , 1.5, 3. ]), 'y': array([0. , 0. , 0. , 1.5, 1.5, 1.5])}, 6)
    """
    if shape is None:
        shape = (5, 10)
    ndim = len(shape)
    if step_sizes is None:
        step_sizes = (1,) * ndim

    if ndim == 3 or len(step_sizes) > 2:
        raise ValueError("Can only create coordinate arrays for 2D maps")

    # Set up as if a 2D map is to be returned
    dy, dx = (1,) * (2 - ndim) + step_sizes
    ny, nx = (1,) * (2 - ndim) + shape
    d = dict()

    # Add coordinate arrays depending on the number of map dimensions
    d["x"] = np.tile(np.arange(nx) * dx, ny).flatten()
    map_size = nx
    if ndim > 1:
        d["y"] = np.sort(np.tile(np.arange(ny) * dy, nx)).flatten()
        map_size *= ny

    return d, map_size



from itertools import chain

from setuptools import find_packages, setup

from orix import __author__, __author_email__, __description__, __name__, __version__

# Projects with optional features for building the documentation and running
# tests. From setuptools:
# https://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-extras-optional-features-with-their-own-dependencies
# fmt: off
extra_feature_requirements = {
    "doc": [
        "ipykernel",  # Used by nbsphinx to execute notebooks
        "memory_profiler",
        # TODO: Remove nbconvert pin once
        #  https://github.com/pyxem/orix/issues/494 is resolved
        "nbconvert                      < 7.14",
        "nbsphinx                       >= 0.7",
        "numpydoc",
        "pydata-sphinx-theme",
        "scikit-image",
        "scikit-learn",
        "sphinx                         >= 3.0.2",
        "sphinx-codeautolink[ipython]",
        "sphinx-copybutton              >= 0.2.5",
        "sphinx-design",
        "sphinx-gallery",
        "sphinxcontrib-bibtex           >= 1.0",
    ],
    "tests": [
        "coverage                       >= 5.0",
        "numpydoc",
        "pytest                         >= 5.4",
        "pytest-cov                     >= 2.8.1",
        "pytest-rerunfailures",
        "pytest-xdist",
    ],
}
extra_feature_requirements["dev"] = [
    "black[jupyter]",
    "isort                              >= 5.10",
    "manifix",
    "outdated",
    "pre-commit                         >= 1.16",
] + list(chain(*list(extra_feature_requirements.values())))
# fmt: on

# Remove the "raw" ReStructuredText directive from the README so we can
# use it as the long_description on PyPI
readme = open("README.rst").read()
readme_split = readme.split("\n")
for i, line in enumerate(readme_split):
    if line == ".. EXCLUDE":
        break
long_description = "\n".join(readme_split[i + 2 :])

setup(
    name=__name__,
    version=str(__version__),
    license="GPLv3",
    url="https://orix.readthedocs.io",
    author=__author__,
    author_email=__author_email__,
    description=__description__,
    long_description=long_description,
    long_description_content_type="text/x-rst",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        (
            "License :: OSI Approved :: GNU General Public License v3 or later "
            "(GPLv3+)"
        ),
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    python_requires=">=3.8",
    packages=find_packages(exclude=["orix/tests"]),
    extras_require=extra_feature_requirements,
    # fmt: off
    install_requires=[
        "dask[array]",
        "diffpy.structure       >= 3.0.2",
        "h5py",
        "matplotlib             >= 3.5",
        "matplotlib-scalebar",
        "numba",
        "numpy",
        "numpy-quaternion",
        "pooch                  >= 0.13",
        "scipy",
        "tqdm",
    ],
    # fmt: on
    package_data={"": ["LICENSE", "README.rst", "readthedocs.yaml"], "orix": ["*.py"]},
)



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Tuple, Union

import numpy as np
from scipy.ndimage import gaussian_filter

from orix.projections import StereographicProjection
from orix.quaternion import Symmetry
from orix.vector import Vector3d


def pole_density_function(
    *args: Union[np.ndarray, Vector3d],
    resolution: float = 1,
    sigma: float = 5,
    weights: Optional[np.ndarray] = None,
    hemisphere: str = "upper",
    symmetry: Optional[Symmetry] = None,
    log: bool = False,
    mrd: bool = True
) -> Tuple[np.ma.MaskedArray, Tuple[np.ndarray, np.ndarray]]:
    """Compute the Pole Density Function (PDF) of vectors in the
    stereographic projection. See :cite:`rohrer2004distribution`.

    If ``symmetry`` is defined then the PDF is folded back into the
    point group fundamental sector and accumulated.

    Parameters
    ----------
    args
        Vector(s), or azimuth and polar angles of the vectors, the
        latter passed as two separate arguments.
    resolution
        The angular resolution of the sampling grid in degrees.
        Default value is 1.
    sigma
        The angular resolution of the applied broadening in degrees.
        Default value is 5.
    weights
        The weights for the individual vectors. Default is ``None``, in
        which case the weight of each vector is 1.
    hemisphere
        Which hemisphere(s) to plot the vectors on, options are
        ``"upper"`` and ``"lower"``. Default is ``"upper"``.
    symmetry
        If provided the PDF is calculated within the fundamental sector
        of the point group symmetry, otherwise the PDF is calculated
        on ``hemisphere``. Default is ``None``.
    log
        If ``True`` the log(PDF) is calculated. Default is ``True``.
    mrd
        If ``True`` the returned PDF is in units of Multiples of Random
        Distribution (MRD), otherwise the units are bin counts. Default
        is ``True``.

    Returns
    -------
    hist
        The computed histogram, shape is (N, M).
    x, y
        Tuple of coordinate grids for the bin edges of ``hist``. The
        units of ``x`` and ``y`` are cartesian coordinates on the
        stereographic projection plane and the shape of both ``x`` and
        ``y`` is (N + 1, M + 1).

    See Also
    --------
    orix.plot.InversePoleFigurePlot.pole_density_function
    orix.plot.StereographicPlot.pole_density_function
    orix.vector.Vector3d.pole_density_function
    """
    from orix.sampling.S2_sampling import _sample_S2_equal_area_coordinates

    hemisphere = hemisphere.lower()

    poles = {"upper": -1, "lower": 1}
    sp = StereographicProjection(poles[hemisphere])

    if len(args) == 1:
        v = args[0]
        if not isinstance(v, Vector3d):
            raise TypeError(
                "If one argument is passed it must be an instance of "
                + "`orix.vector.Vector3d`."
            )
    elif len(args) == 2:
        # azimuth and polar angles
        v = Vector3d.from_polar(*args)
    else:
        raise ValueError(
            "Accepts only one (Vector3d) or two (azimuth, polar) input arguments."
        )

    if symmetry is not None:
        v = v.in_fundamental_sector(symmetry)
        # To help with aliasing after reprojection into point group
        # fundamental sector in the inverse pole figure case, the
        # initial sampling is performed at half the angular resolution
        resolution /= 2

    azimuth, polar, _ = v.to_polar()
    # np.histogram2d expects 1d arrays
    azimuth, polar = np.ravel(azimuth), np.ravel(polar)
    if not azimuth.size:
        raise ValueError("`azimuth` and `polar` angles have 0 size.")

    # Generate equal area mesh on S2
    azimuth_coords, polar_coords = _sample_S2_equal_area_coordinates(
        resolution,
        hemisphere=hemisphere,
        azimuth_endpoint=True,
    )
    azimuth_grid, polar_grid = np.meshgrid(azimuth_coords, polar_coords, indexing="ij")
    # generate histogram in angular space
    hist, *_ = np.histogram2d(
        azimuth,
        polar,
        bins=(azimuth_coords, polar_coords),
        density=False,
        weights=weights,
    )

    # "wrap" along azimuthal axis, "reflect" along polar axis
    mode = ("wrap", "reflect")
    # apply broadening in angular space
    hist = gaussian_filter(hist, sigma / resolution, mode=mode)

    # In the case of inverse pole figure, accumulate all values outside
    # of the point group fundamental sector back into correct bin within
    # fundamental sector
    if symmetry is not None:
        # compute histogram bin centers in azimuth and polar coords
        azimuth_center_grid, polar_center_grid = np.meshgrid(
            azimuth_coords[:-1] + np.diff(azimuth_coords) / 2,
            polar_coords[:-1] + np.diff(polar_coords) / 2,
            indexing="ij",
        )
        v_center_grid = Vector3d.from_polar(
            azimuth=azimuth_center_grid, polar=polar_center_grid
        ).unit
        # fold back in into fundamental sector
        v_center_grid_fs = v_center_grid.in_fundamental_sector(symmetry)
        azimuth_center_fs, polar_center_fs, _ = v_center_grid_fs.to_polar()
        azimuth_center_fs = azimuth_center_fs.ravel()
        polar_center_fs = polar_center_fs.ravel()

        # Generate coorinates with user-defined resolution.
        # When `symmetry` is defined, the initial grid was calculated
        # with `resolution = resolution / 2`
        azimuth_coords_res2, polar_coords_res2 = _sample_S2_equal_area_coordinates(
            2 * resolution,
            hemisphere=hemisphere,
            azimuth_endpoint=True,
        )
        azimuth_res2_grid, polar_res2_grid = np.meshgrid(
            azimuth_coords_res2, polar_coords_res2, indexing="ij"
        )
        v_res2_grid = Vector3d.from_polar(
            azimuth=azimuth_res2_grid, polar=polar_res2_grid
        )

        # calculate histogram values for vectors folded back into
        # fundamental sector
        i = np.digitize(azimuth_center_fs, azimuth_coords_res2[1:-1])
        j = np.digitize(polar_center_fs, polar_coords_res2[1:-1])
        # recompute histogram
        temp = np.zeros((azimuth_coords_res2.size - 1, polar_coords_res2.size - 1))
        # add hist data to new histogram without buffering
        np.add.at(temp, (i, j), hist.ravel())

        # get new histogram bins centers to compute histogram mask
        azimuth_center_res2_grid, polar_center_res2_grid = np.meshgrid(
            azimuth_coords_res2[:-1] + np.ediff1d(azimuth_coords_res2) / 2,
            polar_coords_res2[:-1] + np.ediff1d(polar_coords_res2) / 2,
            indexing="ij",
        )
        v_center_res2_grid = Vector3d.from_polar(
            azimuth=azimuth_center_res2_grid, polar=polar_center_res2_grid
        ).unit

        # compute histogram data array as masked array
        hist = np.ma.array(
            temp, mask=~(v_center_res2_grid <= symmetry.fundamental_sector)
        )
        # calculate bin vertices
        x, y = sp.vector2xy(v_res2_grid)
        x, y = x.reshape(v_res2_grid.shape), y.reshape(v_res2_grid.shape)
    else:
        # all points valid in stereographic projection
        hist = np.ma.array(hist, mask=np.zeros_like(hist, dtype=bool))
        # calculate bin vertices
        v_grid = Vector3d.from_polar(azimuth=azimuth_grid, polar=polar_grid).unit
        x, y = sp.vector2xy(v_grid)
        x, y = x.reshape(v_grid.shape), y.reshape(v_grid.shape)

    # Normalize by the average number of counts per cell on the
    # unit sphere to calculate in terms of Multiples of Random
    # Distribution (MRD). See :cite:`rohrer2004distribution`.
    # as `hist` is a masked array, only valid (unmasked) values are
    # used in this computation
    if mrd:
        hist = hist / hist.mean()

    if log:
        # +1 to avoid taking the log of 0
        hist = np.log(hist + 1)

    return hist, (x, y)



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Projections of :class:`~orix.vector.Vector3d` to various other
representations.
"""

from orix.projections.stereographic import (
    InverseStereographicProjection,
    StereographicProjection,
)

__all__ = [
    "InverseStereographicProjection",
    "StereographicProjection",
]



"""
========
Subplots
========

This example shows how to place different plots in the same figure using orix' various
:mod:`plot types <orix.plot>`, which extend Matplotlib's plot types.

By first creating a blank figure and then using
:meth:`~matplotlib.figure.Figure.add_subplot` to add various orix plot types, we build
up a figure with two rows and two columns showing separate inverse pole figure Z (IPF-Z)
maps of two phases, a pole density function (PDF), and a discrete pole figure (PF).
"""

import matplotlib.pyplot as plt

from orix import data, plot
from orix.vector import Miller

xmap = data.sdss_ferrite_austenite(allow_download=True)
print(xmap)

pg_m3m = xmap.phases[1].point_group.laue
O_au = xmap["austenite"].orientations
O_fe = xmap["ferrite"].orientations

# Orientation colors
ckey_m3m = plot.IPFColorKeyTSL(pg_m3m)
rgb_au = ckey_m3m.orientation2color(O_au)
rgb_fe = ckey_m3m.orientation2color(O_fe)

# Austenite <111> poles in the sample reference frame
t_au = Miller(uvw=[1, 1, 1], phase=xmap.phases["austenite"]).symmetrise(unique=True)
t_au_all = O_au.inv().outer(t_au)

# Ferrite <111> poles in the sample reference frame
t_fe = Miller(uvw=[1, 1, 1], phase=xmap.phases["ferrite"]).symmetrise(unique=True)
t_fe_all = O_fe.inv().outer(t_fe)

# Create figure
fig = plt.figure(figsize=(8, 8))

ax0 = fig.add_subplot(221, projection="plot_map")
ax0.plot_map(xmap["austenite"], rgb_au)
ax0.set_title("Austenite IPF-Z")
ax0.remove_padding()

ax1 = fig.add_subplot(222, projection="stereographic")
ax1.pole_density_function(t_au_all)
ax1.set_labels("X", "Y", None)
ax1.set_title(r"Austenite $\left<111\right>$ PDF")

ax2 = fig.add_subplot(223, projection="plot_map")
ax2.plot_map(xmap["ferrite"], rgb_fe)
ax2.add_overlay(xmap["ferrite"], xmap["ferrite"].dp)
ax2.set_title("Ferrite IPF-Z w/score overlay")
ax2.remove_padding()

ax3 = fig.add_subplot(224, projection="stereographic")
ax3.scatter(t_fe_all)
ax3.set_labels("X", "Y", None)
ax3.set_title(r"Ferrite $\left<111\right>$ PF")

# Place the IPF color key carefully in the center over the other figures
ax_ckey = fig.add_axes(
    [0.45, 0.5, 0.1, 0.1], projection="ipf", symmetry=pg_m3m, zorder=2
)
ax_ckey.plot_ipf_color_key(show_title=False)
ax_ckey.patch.set_facecolor("None")

fig.subplots_adjust(hspace=0, wspace=0.1)



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Optional, Tuple, Union
import warnings

import dask.array as da
from dask.diagnostics import ProgressBar
from diffpy.structure import Structure
from matplotlib.gridspec import SubplotSpec
import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial.transform import Rotation as SciPyRotation

from orix._util import deprecated
from orix.quaternion.misorientation import Misorientation
from orix.quaternion.rotation import Rotation
from orix.quaternion.symmetry import C1, Symmetry, _get_unique_symmetry_elements
from orix.vector import Miller, NeoEuler, Vector3d


class Orientation(Misorientation):
    r"""Orientations represent misorientations away from a reference of
    identity and have only one associated symmetry.

    Orientations :math:`O` support binary subtraction, producing a
    misorientation :math:`M`. That is, to compute the misorientation
    from :math:`O_1` to :math:`O_2`, call :code:`O_2 - O_1`.

    In orix, orientations and misorientations are distinguished from
    rotations only by the inclusion of a notion of symmetry. Consider
    the following example:

    .. image:: /_static/img/orientation.png
       :width: 200px
       :alt: Two objects with two different rotations each. The square,
             with four-fold symmetry, has the same orientation in both
             cases.
       :align: center

    Both objects have undergone the same *rotations* with respect to the
    reference. However, because the square has four-fold symmetry, it is
    indistinguishable in both cases, and hence has the same orientation.
    """

    # -------------------------- Properties -------------------------- #

    @property
    def symmetry(self) -> Symmetry:
        """Symmetry."""
        return self._symmetry[1]

    @symmetry.setter
    def symmetry(self, value: Symmetry):
        if not isinstance(value, Symmetry):
            raise TypeError("Value must be an instance of orix.quaternion.Symmetry.")
        self._symmetry = (C1, value)

    @property
    def unit(self) -> Orientation:
        """Return unit orientations."""
        O = super().unit
        O.symmetry = self.symmetry
        return O

    # ------------------------ Dunder methods ------------------------ #

    def __invert__(self) -> Orientation:
        O = super().__invert__()
        O.symmetry = self.symmetry
        return O

    def __neg__(self) -> Orientation:
        O = super().__neg__()
        O.symmetry = self.symmetry
        return O

    def __repr__(self) -> str:
        """String representation."""
        data = np.array_str(self.data, precision=4, suppress_small=True)
        return f"{self.__class__.__name__} {self.shape} {self.symmetry.name}\n{data}"

    def __sub__(self, other: Orientation) -> Misorientation:
        if isinstance(other, Orientation):
            # Call to Object3d.squeeze() doesn't carry over symmetry
            M = Misorientation(self * ~other).squeeze()
            M.symmetry = (self.symmetry, other.symmetry)
            return M.map_into_symmetry_reduced_zone()
        return NotImplemented

    # ------------------------ Class methods ------------------------- #

    # TODO: Remove use of **kwargs in 1.0
    @classmethod
    def from_euler(
        cls,
        euler: Union[np.ndarray, tuple, list],
        symmetry: Optional[Symmetry] = None,
        direction: str = "lab2crystal",
        degrees: bool = False,
        **kwargs,
    ) -> Orientation:
        """Create orientations from sets of Euler angles
        :cite:`rowenhorst2015consistent`.

        Parameters
        ----------
        euler
            Euler angles in radians (``degrees=False``) or in degrees
            (``degrees=True``) in the Bunge convention.
        symmetry
            Symmetry of orientation(s). If not given (default), no
            symmetry is set.
        direction
            ``"lab2crystal"`` (default) or ``"crystal2lab"``.
            ``"lab2crystal"`` is the Bunge convention. If ``"MTEX"`` is
            provided then the direction is ``"crystal2lab"``.
        degrees
            If ``True``, the given angles are assumed to be in degrees.
            Default is ``False``.

        Returns
        -------
        O
            Orientations.
        """
        O = super().from_euler(euler, direction=direction, degrees=degrees, **kwargs)
        if symmetry:
            O.symmetry = symmetry
        return O

    @classmethod
    def from_align_vectors(
        cls,
        other: Miller,
        initial: Vector3d,
        weights: Optional[np.ndarray] = None,
        return_rmsd: bool = False,
        return_sensitivity: bool = False,
    ) -> Union[
        Orientation,
        Tuple[Orientation, float],
        Tuple[Orientation, np.ndarray],
        Tuple[Orientation, float, np.ndarray],
    ]:
        """Create an estimated orientation to optimally align vectors in
        the crystal and sample reference frames.

        This method wraps
        :meth:`~scipy.spatial.transform.Rotation.align_vectors`. See
        that method for further explanations of parameters and returns.

        Parameters
        ----------
        other
            Crystal directions of shape ``(n,)`` in the crystal
            reference frame.
        initial
            Sample directions of shape ``(n,)`` in the sample reference
            frame.
        weights
            Relative importance of the different vectors.
        return_rmsd
            Whether to return the (weighted) root mean square distance
            between ``other`` and ``initial`` after alignment. Default
            is ``False``.
        return_sensitivity
            Whether to return the sensitivity matrix. Default is
            ``False``.

        Returns
        -------
        estimated_orientation
            Best estimate of the orientation that transforms ``initial``
            to ``other``. The symmetry of the orientation is inferred
            from the point group of the phase of ``other``, if given.
        rmsd
            Returned when ``return_rmsd=True``.
        sensitivity
            Returned when ``return_sensitivity=True``.

        Raises
        ------
        ValueError
            If ``other`` is not a Miller instance.

        Examples
        --------
        >>> from orix.quaternion import Orientation
        >>> from orix.vector import Vector3d, Miller
        >>> from orix.crystal_map import Phase
        >>> t = Miller(uvw=[[0, 1, 0], [1, 0, 0]], phase=Phase(point_group="m-3m"))
        >>> v_sample = Vector3d([[0, -1, 0], [0, 0, 1]])
        >>> O = Orientation.from_align_vectors(t, v_sample)
        >>> O * v_sample
        Vector3d (2,)
        [[0. 1. 0.]
         [1. 0. 0.]]
        """
        if not isinstance(other, Miller):
            raise ValueError(
                f"Argument other must be of type Miller, but has type {type(other)}"
            )

        out = Rotation.from_align_vectors(
            other=other,
            initial=initial,
            weights=weights,
            return_rmsd=return_rmsd,
            return_sensitivity=return_sensitivity,
        )
        out = list(out)
        out[0] = cls(out[0].data)

        try:
            out[0].symmetry = other.phase.point_group
        except (AttributeError, ValueError):
            pass

        return out[0] if len(out) == 1 else tuple(out)

    @classmethod
    def from_matrix(
        cls, matrix: np.ndarray, symmetry: Optional[Symmetry] = None
    ) -> Orientation:
        """Create orientations from orientation matrices
        :cite:`rowenhorst2015consistent`.

        Parameters
        ----------
        matrix
            Array of orientation matrices.
        symmetry
            Symmetry of orientation(s). If not given (default), no
            symmetry is set.

        Returns
        -------
        O
            Orientations.
        """
        O = super().from_matrix(matrix)
        if symmetry:
            O.symmetry = symmetry
        return O

    # TODO: Remove before 0.13.0
    @classmethod
    @deprecated(since="0.12", removal="0.13", alternative="from_axes_angles")
    def from_neo_euler(
        cls, neo_euler: NeoEuler, symmetry: Optional[Symmetry] = None
    ) -> Orientation:
        """Create orientations from a neo-euler (vector) representation.

        Parameters
        ----------
        neo_euler
            Vector parametrization of orientation(s).
        symmetry
            Symmetry of orientation(s). If not given (default), no
            symmetry is set.

        Returns
        -------
        O
            Orientations.
        """
        O = super().from_neo_euler(neo_euler)
        if symmetry:
            O.symmetry = symmetry
        return O

    @classmethod
    def from_axes_angles(
        cls,
        axes: Union[np.ndarray, Vector3d, tuple, list],
        angles: Union[np.ndarray, tuple, list, float],
        symmetry: Optional[Symmetry] = None,
        degrees: bool = False,
    ) -> Orientation:
        """Create orientations from axis-angle pairs
        :cite:`rowenhorst2015consistent`.

        Parameters
        ----------
        axes
            Axes of rotation.
        angles
            Angles of rotation in radians (``degrees=False``) or degrees
            (``degrees=True``).
        symmetry
            Symmetry of orientations. If not given (default), no
            symmetry is set.
        degrees
            If ``True``, the given angles are assumed to be in degrees.
            Default is ``False``.

        Returns
        -------
        O
            Orientation(s).

        Examples
        --------
        >>> from orix.quaternion import Orientation, symmetry
        >>> O = Orientation.from_axes_angles((0, 0, -1), 90, symmetry.Oh, degrees=True)
        >>> O
        Orientation (1,) m-3m
        [[ 0.7071  0.      0.     -0.7071]]
        """
        O = super().from_axes_angles(axes, angles, degrees)
        if symmetry:
            O.symmetry = symmetry
        return O

    @classmethod
    def from_scipy_rotation(
        cls, rotation: SciPyRotation, symmetry: Optional[Symmetry] = None
    ) -> Orientation:
        """Return orientation(s) from
        :class:`scipy.spatial.transform.Rotation`.

        Parameters
        ----------
        rotation
            SciPy rotation(s).
        symmetry
            Crystal symmetry. If not given, the returned orientation(s)
            is given only the identity symmetry operation, *1* (*C1*).

        Returns
        -------
        O
            Orientations.

        Notes
        -----
        The SciPy rotation is inverted to be consistent with the orix
        framework of passive rotations.

        Examples
        --------
        >>> from orix.quaternion import Orientation, symmetry
        >>> from orix.vector import Vector3d
        >>> from scipy.spatial.transform import Rotation as SciPyRotation
        >>> R_scipy = SciPyRotation.from_euler("ZXZ", [90, 0, 0], degrees=True)
        >>> O = Orientation.from_scipy_rotation(R_scipy, symmetry.Oh)
        >>> v = [1, 1, 0]
        >>> R_scipy.apply(v)
        array([-1.,  1.,  0.])
        >>> O * Vector3d(v)
        Vector3d (1,)
        [[ 1. -1.  0.]]
        """
        O = super().from_scipy_rotation(rotation)
        if symmetry:
            O.symmetry = symmetry
        return O

    @classmethod
    def random(
        cls, shape: Union[int, tuple] = 1, symmetry: Optional[Symmetry] = None
    ) -> Orientation:
        """Create random orientations.

        Parameters
        ----------
        shape
            Shape of the orientations.
        symmetry
            Crystal symmetry. If not given, the returned orientation(s)
            is given only the identity symmetry operation, *1* (*C1*).

        Returns
        -------
        O
            Random orientations.
        """
        O = super().random(shape)
        if symmetry:
            O.symmetry = symmetry
        return O

    # --------------------- Other public methods --------------------- #

    def angle_with(self, other: Orientation, degrees: bool = False) -> np.ndarray:
        """Return the smallest symmetry reduced angles of rotation
        transforming the orientations to the other orientations.

        Parameters
        ----------
        other
            Another orientation.
        degrees
            If ``True``, the angles are returned in degrees. Default is
            ``False``.

        Returns
        -------
        angles
            Smallest symmetry reduced angles in radians
            (``degrees=False``) or degrees (``degrees=True``).

        See Also
        --------
        angle_with_outer
        Rotation.angle_with_outer
        """
        dot_products = self.unit.dot(other.unit)
        angles = np.nan_to_num(np.arccos(2 * dot_products**2 - 1))
        if degrees:
            angles = np.rad2deg(angles)
        return angles

    def angle_with_outer(
        self,
        other: Orientation,
        lazy: bool = False,
        chunk_size: int = 20,
        progressbar: bool = True,
        degrees: bool = False,
    ) -> np.ndarray:
        r"""Return the symmetry reduced smallest angle of rotation
        transforming every orientation in this instance to every
        orientation in another instance.

        Parameters
        ----------
        other
            Another orientation.
        lazy
            Whether to perform the computation lazily with Dask. Default
            is ``False``.
        chunk_size
            Number of orientations per axis to include in each iteration
            of the computation. Default is 20. Only applies when
            ``lazy=True``. Increasing this might reduce the computation
            time at the cost of increased memory use.
        progressbar
            Whether to show a progressbar during computation if
            ``lazy=True``. Default is ``True``.
        degrees
            If ``True``, the angles are returned in degrees. Default is
            ``False``.

        Returns
        -------
        angles
            Smallest symmetry reduced angles in radians
            (``degrees=False``) or degrees (``degrees=True``).

        See Also
        --------
        angle_with

        Notes
        -----
        Given two orientations :math:`O_i` and :math:`O_j`, the smallest
        angle is considered as the geodesic distance

        .. math::

            d(O_i, O_j) = \arccos(2(O_i \cdot O_j)^2 - 1),

        where :math:`(O_i \cdot O_j)` is the highest dot product between
        symmetrically equivalent orientations to :math:`O_{i,j}`.

        Examples
        --------
        >>> from orix.quaternion import Orientation, symmetry
        >>> O1 = Orientation.random((5, 3))
        >>> O2 = Orientation.random((6, 2))
        >>> omega1 = O1.angle_with_outer(O2)
        >>> omega1.shape
        (5, 3, 6, 2)
        >>> O1.symmetry = symmetry.Oh
        >>> O2.symmetry = symmetry.Oh
        >>> omega_sym = O1.angle_with_outer(O2)
        >>> np.allclose(omega1.data, omega_sym.data)
        False
        """
        O = self.unit
        if lazy:
            dot_products = O._dot_outer_dask(other, chunk_size=chunk_size)
            # Round because some dot products are slightly above 1
            n_decimals = np.finfo(dot_products.dtype).precision
            dot_products = da.round(dot_products, n_decimals)

            angles_dask = da.arccos(2 * dot_products**2 - 1)
            angles_dask = da.nan_to_num(angles_dask)

            # Create array in memory and overwrite, chunk by chunk
            angles = np.zeros(angles_dask.shape)
            if progressbar:
                with ProgressBar():
                    da.store(sources=angles_dask, targets=angles)
            else:
                da.store(sources=angles_dask, targets=angles)
        else:
            dot_products = O.dot_outer(other)
            angles = np.arccos(2 * dot_products**2 - 1)
            angles = np.nan_to_num(angles)

        if degrees:
            angles = np.rad2deg(angles)

        return angles

    def get_distance_matrix(
        self,
        lazy: bool = False,
        chunk_size: int = 20,
        progressbar: bool = True,
        degrees: bool = False,
    ) -> np.ndarray:
        r"""Return the symmetry reduced smallest angle of rotation
        transforming all these orientations to all the other
        orientations :cite:`johnstone2020density`.

        Parameters
        ----------
        lazy
            Whether to perform the computation lazily with Dask. Default
            is ``False``.
        chunk_size
            Number of orientations per axis to include in each iteration
            of the computation. Default is 20. Only applies when
            ``lazy=True``. Increasing this might reduce the computation
            time at the cost of increased memory use.
        progressbar
            Whether to show a progressbar during computation if
            ``lazy=True``. Default is ``True``.
        degrees
            If ``True``, the angles are returned in degrees. Default is
            ``False``.

        Returns
        -------
        angles
            Symmetry reduced angles in radians (``degrees=False``) or
            degrees (``degrees=True``).

        Notes
        -----
        Given two orientations :math:`g_i` and :math:`g_j`, the smallest
        angle is considered as the geodesic distance

        .. math::

            d(O_i, O_j) = \arccos(2(O_i \cdot O_j)^2 - 1),

        where :math:`(O_i \cdot O_j)` is the highest dot product between
        symmetrically equivalent orientations to :math:`O_{i,j}`.
        """
        angles = self.angle_with_outer(
            self,
            lazy=lazy,
            chunk_size=chunk_size,
            progressbar=progressbar,
            degrees=degrees,
        )
        return angles

    def dot(self, other: Orientation) -> np.ndarray:
        """Return the symmetry reduced dot products of the orientations
        and the other orientations.

        Parameters
        ----------
        other
            Other orientations.

        Returns
        -------
        highest_dot_products
            Symmetry reduced dot products.

        See Also
        --------
        dot_outer

        Examples
        --------
        >>> from orix.quaternion import Orientation, symmetry
        >>> O1 = Orientation.from_axes_angles([0, 0, 1], [0, 45], symmetry.Oh, degrees=True)
        >>> O2 = Orientation.from_axes_angles([0, 0, 1], [45, 90], symmetry.Oh, degrees=True)
        >>> O1.dot(O2)
        array([0.92387953, 0.92387953])
        """
        symmetry = _get_unique_symmetry_elements(self.symmetry, other.symmetry)
        M = other * ~self
        all_dot_products = Rotation(M).dot_outer(symmetry)
        highest_dot_products = np.max(all_dot_products, axis=-1)
        return highest_dot_products

    def dot_outer(self, other: Orientation) -> np.ndarray:
        """Return the symmetry reduced dot products of all orientations
        to all other orientations.

        Parameters
        ----------
        other
            Other orientations.

        Returns
        -------
        highest_dot_products
            Symmetry reduced dot products.

        See Also
        --------
        dot

        Examples
        --------
        >>> from orix.quaternion import Orientation, symmetry
        >>> O1 = Orientation.from_axes_angles([0, 0, 1], [0, 45], symmetry.Oh, degrees=True)
        >>> O2 = Orientation.from_axes_angles([0, 0, 1], [45, 90], symmetry.Oh, degrees=True)
        >>> O1.dot_outer(O2)
        array([[0.92387953, 1.        ],
               [1.        , 0.92387953]])
        """
        symmetry = _get_unique_symmetry_elements(self.symmetry, other.symmetry)
        M = other.outer(~self)
        all_dot_products = Rotation(M).dot_outer(symmetry)
        highest_dot_products = np.max(all_dot_products, axis=-1)
        # need to return axes order so that self is first
        order = tuple(range(self.ndim, self.ndim + other.ndim)) + tuple(
            range(self.ndim)
        )
        return highest_dot_products.transpose(*order)

    def plot_unit_cell(
        self,
        c: str = "tab:blue",
        return_figure: bool = False,
        axes_length: float = 0.5,
        structure: Optional[Structure] = None,
        crystal_axes_loc: str = "origin",
        **arrow_kwargs,
    ) -> plt.Figure:
        """Plot the unit cell orientation, showing the sample and
        crystal reference frames.

        Parameters
        ----------
        c
            Unit cell edge color.
        return_figure
            Return the plotted figure.
        axes_length
            Length of the reference axes in Angstroms, by default 0.5.
        structure
            Structure of the unit cell, only orthorhombic lattices are
            currently supported. If not given, a cubic unit cell with a
            lattice parameter of 2 Angstroms will be plotted.
        crystal_axes_loc
            Plot the crystal reference frame axes at the ``"origin"``
            (default) or ``"center"`` of the plotted cell.
        **arrow_kwargs
            Keyword arguments passed to
            :class:`matplotlib.patches.FancyArrowPatch`, for example
            ``"arrowstyle"``.

        Returns
        -------
        fig
            The plotted figure, returned if ``return_figure=True``.

        Raises
        ------
        ValueError
            If :attr:`size` > 1.
        """
        if self.size > 1:
            raise ValueError("Can only plot a single unit cell, so *size* must be 1")

        from orix.plot.unit_cell_plot import _plot_unit_cell

        fig = _plot_unit_cell(
            self,
            c=c,
            axes_length=axes_length,
            structure=structure,
            crystal_axes_loc=crystal_axes_loc,
            **arrow_kwargs,
        )

        if return_figure:
            return fig

    def in_euler_fundamental_region(self) -> np.ndarray:
        """Euler angles in the fundamental Euler region of the proper
        subgroup.

        The Euler angle ranges of each proper subgroup are given in
        :attr:`~orix.quaternion.Symmetry.euler_fundamental_region`.

        From the procedure in MTEX' :code:`quaternion.project2EulerFR`.

        Returns
        -------
        euler_in_region
            Euler angles in radians.
        """
        pg = self.symmetry.proper_subgroup

        # Symmetrize every orientation by operations of the proper
        # subgroup different from rotation about the c-axis
        O = pg._special_rotation.outer(self)

        alpha, beta, gamma = O.to_euler().T
        gamma = np.mod(gamma, 2 * np.pi / pg._primary_axis_order)

        # Find the first triplet among the symmetrically equivalent ones
        # inside the fundamental region
        max_alpha, max_beta, max_gamma = np.radians(pg.euler_fundamental_region)
        is_inside = (alpha <= max_alpha) * (beta <= max_beta) * (gamma <= max_gamma)
        first_nonzero = np.argmax(is_inside, axis=1)

        euler_in_region = np.column_stack(
            (
                np.choose(first_nonzero, alpha.T),
                np.choose(first_nonzero, beta.T),
                np.choose(first_nonzero, gamma.T),
            )
        )

        return euler_in_region

    def scatter(
        self,
        projection: str = "axangle",
        figure: Optional[plt.Figure] = None,
        position: Union[int, Tuple[int], SubplotSpec] = (1, 1, 1),
        return_figure: bool = False,
        wireframe_kwargs: Optional[dict] = None,
        size: Optional[int] = None,
        direction: Optional[Vector3d] = None,
        figure_kwargs: Optional[dict] = None,
        **kwargs,
    ) -> plt.Figure:
        """Plot orientations in axis-angle space, the Rodrigues
        fundamental zone, or an inverse pole figure (IPF) given a sample
        direction.

        Parameters
        ----------
        projection
            Which orientation space to plot orientations in, either
            "axangle" (default), "rodrigues" or "ipf" (inverse pole
            figure).
        figure
            If given, a new plot axis :class:`~orix.plot.AxAnglePlot` or
            :class:`~orix.plot.RodriguesPlot` is added to the figure in
            the position specified by `position`. If not given, a new
            figure is created.
        position
            Where to add the new plot axis. 121 or (1, 2, 1) places it
            in the first of two positions in a grid of 1 row and 2
            columns. See :meth:`~matplotlib.figure.Figure.add_subplot`
            for further details. Default is (1, 1, 1).
        return_figure
            Whether to return the figure. Default is False.
        wireframe_kwargs
            Keyword arguments passed to
            :meth:`orix.plot.AxAnglePlot.plot_wireframe` or
            :meth:`orix.plot.RodriguesPlot.plot_wireframe`.
        size
            If not given, all orientations are plotted. If given, a
            random sample of this `size` of the orientations is plotted.
        direction
            Sample direction to plot with respect to crystal directions.
            If not given, the out of plane direction, sample Z, is used.
            Only used when plotting IPF(s).
        figure_kwargs
            Dictionary of keyword arguments passed to
            :func:`matplotlib.pyplot.figure` if `figure` is not given.
        **kwargs
            Keyword arguments passed to
            :meth:`orix.plot.AxAnglePlot.scatter`,
            :meth:`orix.plot.RodriguesPlot.scatter`, or
            :meth:`orix.plot.InversePoleFigurePlot.scatter`.

        Returns
        -------
        figure
            Figure with the added plot axis, if ``return_figure=True``.

        See Also
        --------
        orix.plot.AxAnglePlot, orix.plot.RodriguesPlot,
        orix.plot.InversePoleFigurePlot
        """
        if projection.lower() != "ipf":
            figure = super().scatter(
                projection=projection,
                figure=figure,
                position=position,
                return_figure=return_figure,
                wireframe_kwargs=wireframe_kwargs,
                size=size,
                figure_kwargs=figure_kwargs,
                **kwargs,
            )
        else:
            from orix.plot.inverse_pole_figure_plot import (
                _setup_inverse_pole_figure_plot,
            )

            if figure is None:
                # Determine which hemisphere(s) to show
                symmetry = self.symmetry
                sector = symmetry.fundamental_sector
                if np.any(sector.vertices.polar > np.pi / 2):
                    hemisphere = "both"
                else:
                    hemisphere = "upper"

                figure, axes = _setup_inverse_pole_figure_plot(
                    symmetry=symmetry,
                    direction=direction,
                    hemisphere=hemisphere,
                    figure_kwargs=figure_kwargs,
                )
            else:
                axes = np.asarray(figure.axes)

            for ax in axes:
                ax.scatter(self, **kwargs)

        if return_figure:
            return figure

    def inv(self) -> Orientation:
        r"""Return the inverse orientations :math:`O^{-1}`."""
        return self.__invert__()

    # -------------------- Other private methods --------------------- #

    def _dot_outer_dask(self, other: Orientation, chunk_size: int = 20) -> da.Array:
        """Symmetry reduced dot product of every orientation in this
        instance to every orientation in another instance, returned as a
        Dask array.

        Parameters
        ----------
        other
        chunk_size
            Number of orientations per axis to include in each iteration
            of the computation. Default is 20. Increasing this might
            reduce the computation time at the cost of increased memory
            use.

        Returns
        -------
        highest_dot_product

        Notes
        -----
        To read the dot products array `dparr` into memory, do
        `dp = dparr.compute()`.
        """
        symmetry = _get_unique_symmetry_elements(self.symmetry, other.symmetry)
        M = other._outer_dask(~self, chunk_size=chunk_size)

        # Summation subscripts
        str1 = "abcdefghijklmnopqrstuvwxy"[: M.ndim]
        str2 = "z" + str1[-1]  # Last elements have shape (4,)
        sum_over = f"{str1},{str2}->{str1[:-1] + str2[0]}"

        warnings.filterwarnings("ignore", category=da.PerformanceWarning)

        all_dot_products = da.einsum(sum_over, M, symmetry.data)
        highest_dot_product = da.max(abs(all_dot_products), axis=-1)

        return highest_dot_product



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Generation of grids on *S2* (vectors) or *SO(3)* (rotations)."""

from orix.sampling.S2_sampling import (
    sample_S2,
    sample_S2_cube_mesh,
    sample_S2_equal_area_mesh,
    sample_S2_hexagonal_mesh,
    sample_S2_icosahedral_mesh,
    sample_S2_random_mesh,
    sample_S2_uv_mesh,
)
from orix.sampling.S2_sampling import sampling_methods as sample_S2_methods
from orix.sampling.SO3_sampling import uniform_SO3_sample
from orix.sampling.sample_generators import (
    get_sample_fundamental,
    get_sample_local,
    get_sample_reduced_fundamental,
)

__all__ = [
    "get_sample_fundamental",
    "get_sample_reduced_fundamental",
    "get_sample_local",
    "sample_S2",
    "sample_S2_methods",
    "uniform_SO3_sample",
    "sample_S2_cube_mesh",
    "sample_S2_equal_area_mesh",
    "sample_S2_hexagonal_mesh",
    "sample_S2_icosahedral_mesh",
    "sample_S2_random_mesh",
    "sample_S2_uv_mesh",
]



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Optional

from orix.quaternion import Symmetry
from orix.vector import Vector3d


class IPFColorKey(abc.ABC):
    """Assign colors to crystal directions rotated by crystal
    orientations and projected into an inverse pole figure.

    This is an abstract class defining properties and methods required
    in derived classes.

    Parameters
    ----------
    symmetry
    direction
    """

    def __init__(self, symmetry: Symmetry, direction: Optional[Vector3d] = None):
        self.symmetry = symmetry
        if direction is None:
            direction = Vector3d.zvector()
        self.direction = direction

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}, symmetry: {self.symmetry.name}, "
            f"direction: {self.direction.data.squeeze()}"
        )

    @property
    @abc.abstractmethod
    def direction_color_key(self):
        return NotImplemented  # pragma: no cover

    @abc.abstractmethod
    def orientation2color(self, *args, **kwargs):
        return NotImplemented  # pragma: no cover

    @abc.abstractmethod
    def plot(self, *args, **kwargs):
        return NotImplemented  # pragma: no cover



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from orix import sampling
from orix.sampling import S2_sampling
from orix.vector import Vector3d


class TestS2Sampling:
    @pytest.mark.parametrize("resolution, size", [(3, 7082), (4, 3962)])
    def test_uv_mesh(self, resolution, size):
        v1 = S2_sampling.sample_S2_uv_mesh(resolution)
        assert v1.size == size
        assert np.allclose(v1.mean().data, [0, 0, 0])
        # test correct number of polar levels (z)
        expected_num_polar_z = int(np.ceil(180 / resolution)) + 1
        z_rounded = v1.z.round(6)
        assert np.unique(z_rounded).size == expected_num_polar_z
        # test no duplicate azimuthal values, get a central elevation
        z_unique = np.unique(z_rounded)
        expected_num_azimuth = int(np.ceil(360 / resolution))
        v1_elev = v1[np.isclose(z_rounded, z_unique[z_unique.size // 2])]
        assert np.unique(np.arctan2(v1_elev.y, v1_elev.x)).size == expected_num_azimuth
        # check only one pole at north and south pole
        assert np.isclose(v1.z, 1).sum() == 1
        assert np.isclose(v1.z, -1).sum() == 1

    @pytest.mark.parametrize(
        "hemisphere, min_polar, max_polar",
        [("upper", 0, np.pi / 2), ("lower", np.pi / 2, np.pi), ("both", 0, np.pi)],
    )
    def test_uv_mesh_coordinate_arrays_hemisphere(
        self, hemisphere, min_polar, max_polar
    ):
        azi, polar = S2_sampling._sample_S2_uv_mesh_coordinates(
            10, hemisphere=hemisphere, azimuth_endpoint=False
        )
        assert isinstance(azi, np.ndarray)
        assert isinstance(polar, np.ndarray)
        assert azi.ndim == 1
        assert polar.ndim == 1
        assert polar.min() == min_polar
        assert polar.max() == max_polar
        assert azi.min() == 0
        assert azi.max() < 2 * np.pi
        assert azi.size == 36
        azi2, _ = S2_sampling._sample_S2_equal_area_coordinates(
            10, hemisphere=hemisphere, azimuth_endpoint=True
        )
        assert azi2.max() == 2 * np.pi

    def test_uv_mesh_raises(self):
        with pytest.raises(ValueError, match="Hemisphere must be one of "):
            sampling.sample_S2_uv_mesh(10, hemisphere="test")

        with pytest.raises(
            ValueError, match="Offset is a fractional value of the angular step size "
        ):
            sampling.sample_S2_uv_mesh(10, hemisphere="upper", offset=100)

    @pytest.mark.parametrize(
        "grid_type, resolution, size",
        [
            ("spherified_corner", 3, 8666),
            ("normalized", 4, 5402),
            ("spherified_edge", 5, 1946),
        ],
    )
    def test_cube_mesh(self, grid_type, resolution, size):
        v1 = S2_sampling.sample_S2_cube_mesh(resolution, grid_type=grid_type)
        assert v1.size == size
        assert np.allclose(v1.mean().data, [0, 0, 0])
        assert v1.unique().size == size

    def test_cube_mesh_raises(self):
        with pytest.raises(ValueError, match="The `grid_type` hexagonal"):
            _ = sampling.sample_S2_cube_mesh(2, "hexagonal")

    @pytest.mark.parametrize(
        "hemisphere, min_polar, max_polar",
        [("upper", 0, np.pi / 2), ("lower", np.pi / 2, np.pi), ("both", 0, np.pi)],
    )
    def test_equal_area_coordinates(self, hemisphere, min_polar, max_polar):
        azi, polar = S2_sampling._sample_S2_equal_area_coordinates(
            10, hemisphere=hemisphere, azimuth_endpoint=False
        )
        assert isinstance(azi, np.ndarray)
        assert isinstance(polar, np.ndarray)
        assert azi.ndim == 1
        assert polar.ndim == 1
        assert polar.min() == min_polar
        assert polar.max() == max_polar
        assert azi.min() == 0
        assert azi.max() < 2 * np.pi
        assert azi.size == 36
        azi2, _ = S2_sampling._sample_S2_equal_area_coordinates(
            10, hemisphere=hemisphere, azimuth_endpoint=True
        )
        assert azi2.max() == 2 * np.pi

    def test_equal_area_coordinates_ranges(self):
        azi, polar = S2_sampling._sample_S2_equal_area_coordinates(
            10,
            azimuth_range=(-0.1, 2 * np.pi + 0.1),
            polar_range=(-0.1, np.pi + 0.1),
            azimuth_endpoint=True,
        )

        # code fixes wrap around in ranges
        assert azi.min() == 0
        assert np.allclose(azi.max(), 2 * np.pi)
        assert polar.min() == 0
        assert np.allclose(polar.max(), np.pi)

    def test_equal_area_coordinates_ranges_raises(self):
        with pytest.raises(ValueError, match="`azimuth_range` requires values "):
            S2_sampling._sample_S2_equal_area_coordinates(10, azimuth_range=(1, -1))

        with pytest.raises(ValueError, match="`polar_range` requires values "):
            S2_sampling._sample_S2_equal_area_coordinates(10, polar_range=(1, -1))

    def test_equal_are_coordinates_hemisphere_raises(self):
        with pytest.raises(ValueError, match="`hemisphere` must be one of "):
            S2_sampling._sample_S2_equal_area_coordinates(10, hemisphere="test")

    def test_equal_area_mesh(self):
        v = sampling.sample_S2_equal_area_mesh(
            10, hemisphere="both", remove_pole_duplicates=True
        )
        assert isinstance(v, Vector3d)
        azi, polar, _ = v.to_polar()
        assert polar.ndim == 1
        assert azi.shape == polar.shape
        assert np.count_nonzero(polar == 0) == 1
        assert np.count_nonzero(polar == np.pi) == 1
        v2 = sampling.sample_S2_equal_area_mesh(
            10, hemisphere="both", remove_pole_duplicates=False
        )
        azi2, polar2, _ = v2.to_polar()
        assert polar2.ndim == 2
        assert azi2.shape == polar2.shape
        assert np.count_nonzero(polar2 == 0) > 1
        assert np.count_nonzero(polar2 == np.pi) > 1

    def test_equal_area_mesh_raises(self):
        with pytest.raises(ValueError, match="`hemisphere` must be one of "):
            sampling.sample_S2_equal_area_mesh(10, hemisphere="test")

    def test_random_sphere_mesh(self):
        grid = S2_sampling.sample_S2_random_mesh(1).data
        assert grid.shape[0] == 10313
        assert grid.shape[1] == 3

    def test_seed_for_random_sphere_mesh(self):
        grid_7 = S2_sampling.sample_S2_random_mesh(resolution=3, seed=7)
        grid_7_again = S2_sampling.sample_S2_random_mesh(resolution=3, seed=7)
        grid_8 = S2_sampling.sample_S2_random_mesh(resolution=3, seed=8)
        assert np.allclose(grid_7.data, grid_7_again.data)
        assert not np.allclose(grid_7.data, grid_8.data)

    def test_hexagonal_mesh(self):
        grid = S2_sampling.sample_S2_hexagonal_mesh(5).data
        assert grid.shape == (3458, 3)

    def test_icosahedral_grid(self):
        grid = S2_sampling.sample_S2_icosahedral_mesh(10).data
        assert grid.shape[0] == 642
        assert grid.shape[1] == 3
        np.testing.assert_almost_equal(np.sum(grid), 0)
        grid_unique = np.unique(grid, axis=0)
        assert grid.shape[0] == grid_unique.shape[0]

    def test_sample_s2(self):
        resolution = 10
        grid = S2_sampling.sample_S2(resolution, method="hexagonal")
        grid_2 = S2_sampling.sample_S2_hexagonal_mesh(resolution)
        assert np.allclose(grid.data, grid_2.data)

    def test_sample_s2_nonexistent(self):
        with pytest.raises(NotImplementedError):
            _ = S2_sampling.sample_S2(10, "batman")



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Any, Optional, Tuple, Union
import warnings

import dask.array as da
from dask.diagnostics import ProgressBar
import numpy as np
import quaternion
from scipy.spatial.transform import Rotation as SciPyRotation

from orix._base import Object3d
from orix._util import deprecated, deprecated_argument
from orix.quaternion import _conversions
from orix.vector import AxAngle, Homochoric, Miller, Rodrigues, Vector3d

# Used to round values below 1e-16 to zero
_FLOAT_EPS = np.finfo(float).eps


class Quaternion(Object3d):
    r"""Quaternions.

    Quaternions support the following mathematical operations:

    * Unary negation.
    * Inversion (conjugation).
    * Normalization to obtain unit quaternions.
    * Multiplication with other quaternions and vectors.

    A quaternion :math:`Q` is defined as a four-component number of the
    form :math:`Q = a + ib + jc + kd`, where the imaginary units
    :math:`(i, j, k)` satisfy the following relations:

    .. math::

        i^2 = j^2 = k^2 = -1;

        ij = -ji = k; jk = -kj = i; ki = -ik = j.

    In orix, quaternions are stored with the scalar part first followed
    by the vector part, denoted :math:`Q = (a, b, c, d)`.

    Multiplication of two quaternions :math:`Q_1 = (a_1, b_1, c_1, d_1)`
    and :math:`Q_2 = (a_2, b_2, c_2, d_2)` with
    :math:`Q_3 = q1 \cdot q2 = (a_3, b_3, c_3, d_3)` is performed as:

    .. math::

        a_3 = a_1 \cdot a_2 - b_1 \cdot b_2 - c_1 \cdot c_2 - d_1 \cdot d_2

        b_3 = a_1 \cdot b_2 + b_1 \cdot a_2 + c_1 \cdot d_2 - d_1 \cdot c_2

        c_3 = a_1 \cdot c_2 - b_1 \cdot d_2 + c_1 \cdot a_2 + d_1 \cdot b_2

        d_3 = a_1 \cdot d_2 + b_1 \cdot c_2 - c_1 \cdot b_2 + d_1 \cdot a_2

    Rotation of a 3D vector :math:`v = (x, y, z)` by a quaternion is
    performed as :math:`v' = Q \cdot v \cdot Q^{-1}`. Written out:

    .. math::

        v'_x = x(a^2 + b^2 - c^2 - d^2) + 2[z(a \cdot c + b \cdot d) + y(b \cdot c - a \cdot d)]

        v'_y = y(a^2 - b^2 + c^2 - d^2) + 2[x(a \cdot d + b \cdot c) + z(c \cdot d - a \cdot b)]

        v'_z = z(a^2 - b^2 - c^2 + d^2) + 2[y(a \cdot b + c \cdot d) + x(b \cdot d - a \cdot c)]

    The norm of a quaternion is defined as

    .. math::

        |Q| = \sqrt{a^2 + b^2 + c^2 + d^2}.

    Unit quaternions have a norm of :math:`|Q| = 1` and can always be written on the
    form

    .. math::

        Q = \cos\frac{\omega}{2} + \sin\frac{\omega}{2}(bi + cj + dk),

    where :math:`(b, c, d)` are the direction cosines of the rotation
    axis unit vector :math:`\hat{\mathbf{n}}`. The scalar part
    :math:`a = \cos\frac{\omega}{2}` will always be positive or 0 for
    rotations with rotation angle :math:`\omega = \pi`.

    Conventions:

    1. Right-handed Cartesian reference frames.
    2. Rotation angles :math:`\omega` are taken to be positive for a
       counter-clockwise rotation when viewing from the end point of the
       rotation axis unit vector :math:`\hat{\mathbf{n}}` towards the
       origin.
    3. Rotations are *interpreted* in the passive sense. This means that
       rotations are interpreted as basis transformations of reference
       frames, with vectors fixed in space.
    4. Euler angle triplets are implemented using the Bunge convention,
       with angular ranges as :math:`[0, 2\pi]`, :math:`[0, \pi]`, and
       :math:`[0, 2\pi]`.
    5. Rotation angles :math:`\omega` are limited to :math:`[0, \pi]`.
    """

    dim = 4

    # -------------------------- Properties -------------------------- #

    @property
    def a(self) -> np.ndarray:
        """Return or set the scalar quaternion component.

        Parameters
        ----------
        value : numpy.ndarray
            Scalar quaternion component.
        """
        return self.data[..., 0]

    @a.setter
    def a(self, value: np.ndarray):
        """Set the scalar quaternion component."""
        self.data[..., 0] = value

    @property
    def b(self) -> np.ndarray:
        """Return or set the first vector quaternion component.

        Parameters
        ----------
        value : numpy.ndarray
            First vector quaternion component.
        """
        return self.data[..., 1]

    @b.setter
    def b(self, value: np.ndarray):
        """Set the first vector quaternion component."""
        self.data[..., 1] = value

    @property
    def c(self) -> np.ndarray:
        """Return or set the second vector quaternion component.

        Parameters
        ----------
        value : numpy.ndarray
            Second vector quaternion component.
        """
        return self.data[..., 2]

    @c.setter
    def c(self, value: np.ndarray):
        """Set the second vector quaternion component."""
        self.data[..., 2] = value

    @property
    def d(self) -> np.ndarray:
        """Return or set the third vector quaternion component.

        Parameters
        ----------
        value : numpy.ndarray
            Third vector quaternion component.
        """
        return self.data[..., 3]

    @d.setter
    def d(self, value: np.ndarray):
        """Set the third vector quaternion component."""
        self.data[..., 3] = value

    @property
    def axis(self) -> Vector3d:
        r"""Return the axis of rotation
        :math:`\hat{\mathbf{n}} = (b, c, d)`.
        """
        axis = Vector3d(np.stack((self.b, self.c, self.d), axis=-1))
        a_is_zero = self.a < -1e-6
        axis[a_is_zero] = -axis[a_is_zero]
        norm_is_zero = axis.norm == 0
        axis[norm_is_zero] = Vector3d.zvector() * np.sign(self.a[norm_is_zero])
        axis.data /= axis.norm[..., np.newaxis]
        return axis

    @property
    def angle(self) -> np.ndarray:
        r"""Return the angle of rotation :math:`\omega = 2\arccos{|a|}`."""
        return 2 * np.nan_to_num(np.arccos(np.abs(self.a)))

    @property
    def antipodal(self) -> Quaternion:
        """Return the quaternion and its antipodal."""
        return self.__class__(np.stack([self.data, -self.data]))

    @property
    def conj(self) -> Quaternion:
        r"""Return the conjugate of the quaternion
        :math:`Q^* = a - bi - cj - dk`.
        """
        Q = quaternion.from_float_array(self.data).conj()
        return self.__class__(quaternion.as_float_array(Q))

    # ------------------------ Dunder methods ------------------------ #

    def __invert__(self) -> Quaternion:
        return self.__class__(self.conj.data / (self.norm**2)[..., np.newaxis])

    def __mul__(self, other: Union[Quaternion, Vector3d]):
        if isinstance(other, Quaternion):
            Q1 = quaternion.from_float_array(self.data)
            Q2 = quaternion.from_float_array(other.data)
            return other.__class__(quaternion.as_float_array(Q1 * Q2))
        elif isinstance(other, Vector3d):
            # check broadcast shape is correct before calculation, as
            # quaternion.rotat_vectors will perform outer product
            # this keeps current __mul__ broadcast behaviour
            Q1 = quaternion.from_float_array(self.data)
            v = quaternion.as_vector_part(
                (Q1 * quaternion.from_vector_part(other.data)) * ~Q1
            )
            if isinstance(other, Miller):
                m = other.__class__(xyz=v, phase=other.phase)
                m.coordinate_format = other.coordinate_format
                return m
            else:
                return other.__class__(v)
        return NotImplemented

    def __neg__(self) -> Quaternion:
        return self.__class__(-self.data)

    def __eq__(self, other: Union[Any, Quaternion]) -> bool:
        """Check if quaternions have equal shapes and values."""
        if (
            isinstance(other, Quaternion)
            and self.shape == other.shape
            and np.allclose(self.data, other.data)
        ):
            return True
        else:
            return False

    # ------------------------ Class methods ------------------------- #

    # TODO: Remove before 0.13.0
    @classmethod
    @deprecated(since="0.12", removal="0.13", alternative="from_axes_angles")
    def from_neo_euler(cls, neo_euler: "NeoEuler") -> Quaternion:
        """Create unit quaternion(s) from a neo-euler (vector)
        representation.

        Parameters
        ----------
        neo_euler
            Vector parametrization of quaternions.

        Returns
        -------
        Q
            Unit quaternion(s).
        """
        s = np.sin(neo_euler.angle / 2)
        a = np.cos(neo_euler.angle / 2)
        b = s * neo_euler.axis.x
        c = s * neo_euler.axis.y
        d = s * neo_euler.axis.z
        Q = cls(np.stack([a, b, c, d], axis=-1)).unit
        return Q

    @classmethod
    def from_axes_angles(
        cls,
        axes: Union[np.ndarray, Vector3d, tuple, list],
        angles: Union[np.ndarray, tuple, list, float],
        degrees: bool = False,
    ) -> Quaternion:
        r"""Create unit quaternions from axis-angle pairs
        :math:`(\hat{\mathbf{n}}, \omega)`
        :cite:`rowenhorst2015consistent`.

        Parameters
        ----------
        axes
            Axes of rotation :math:`\hat{\mathbf{n}}`.
        angles
            Angles of rotation :math:`\omega` in radians
            (``degrees=False``) or degrees (``degrees=True``).
        degrees
            If ``True``, the given angles are assumed to be in degrees.
            Default is ``False``.

        Returns
        -------
        Q
            Unit quaternions.

        Examples
        --------
        >>> from orix.quaternion import Quaternion
        >>> Q = Quaternion.from_axes_angles((0, 0, -1), 90, degrees=True)
        >>> Q
        Quaternion (1,)
        [[ 0.7071  0.      0.     -0.7071]]

        See Also
        --------
        from_rodrigues, from_homochoric
        """
        if np.size(axes) == 0:
            return cls.empty()

        axes = Vector3d(axes).unit.data
        angles = np.array(angles)
        if degrees:
            angles = np.deg2rad(angles)

        Q = _conversions.ax2qu(axes, angles)
        Q = cls(Q)
        Q = Q.unit

        return Q

    @classmethod
    def from_homochoric(
        cls,
        ho: Union[Vector3d, Homochoric, np.ndarray, tuple, list],
    ) -> Quaternion:
        r"""Create unit quaternions from homochoric vectors
        :math:`\mathbf{h}` :cite:`rowenhorst2015consistent`.

        Parameters
        ----------
        ho
            Homochoric vectors parallel to the axes of rotation with
            lengths equal to
            :math:`\left[\frac{3}{4}\cdot(\omega - \sin\omega)\right]^{1/3}`,
            where :math:`\omega` is the angle of rotation.

        Returns
        -------
        Q
            Unit quaternions.

        See Also
        --------
        from_axes_angles, from_rodrigues
        """
        if np.size(ho) == 0:
            return cls.empty()

        if isinstance(ho, Vector3d):
            ho = ho.data
        else:
            ho = np.atleast_2d(ho)
            if ho.shape[-1] != 3:
                raise ValueError("Final dimension of vector array must be 3.")

        shape = ho.shape[:-1]
        ho = ho.reshape(-1, 3)

        ax = _conversions.ho2ax(ho)
        Q = _conversions.ax2qu(ax[:, :3], ax[:, 3])
        Q = Q.reshape(*shape, 4)
        Q = cls(Q)
        Q = Q.unit

        return Q

    @classmethod
    def from_rodrigues(
        cls,
        ro: Union[np.ndarray, Vector3d, tuple, list],
        angles: Union[np.ndarray, tuple, list, float, None] = None,
    ) -> Quaternion:
        r"""Create unit quaternions from three-component Rodrigues
        vectors :math:`\hat{\mathbf{n}}` or four-component
        Rodrigues-Frank vectors :math:`\mathbf{\rho}`
        :cite:`rowenhorst2015consistent`.

        Parameters
        ----------
        ro
            Rodrigues vectors :math:`\hat{\mathbf{n}}` of three
            components. These are the components of the Rodrigues-Frank
            vectors :math:`\mathbf{\rho}` if the angles :math:`\omega`
            are passed.
        angles
            Angles :math:`\omega` of the Rodrigues-Frank vectors
            :math:`\mathbf{\rho}`, one per vector. If these are not
            passed, ``ro`` are the Rodrigues vectors
            :math:`\hat{\mathbf{n}}`.

        Returns
        -------
        Q
            Unit quaternions.

        See Also
        --------
        from_axes_angles, from_homochoric

        Notes
        -----
        The Rodrigues-Frank vector :math:`\mathbf{\rho}` is defined as

        .. math::

            \mathbf{\rho} = \hat{\mathbf{n}}\tan\frac{\omega}{2}.

        If the vector length is :math:`\rho = |\mathbf{\rho}|`, the
        angle is given by

        .. math::

            \omega = 2\arctan\rho.

        O. Rodrigues's 1840 vector description was popularized by F. C.
        Frank due to its useful rectilinear mapping of fundamental
        zones, as is well-demonstrated in :cite:`frank1988orientation`.
        However, the length of these vectors, and thus their accuracy,
        scales with :math:`\tan\omega/2`. Additionally, two-fold
        rotations produce vectors of infinite length. Thus, Frank and
        others introduced the Rodrigues-Frank vector of length 4,
        consisting of a unit vector followed by the scaling factor
        :math:`\tan\omega/2`. This is better suited for storing data
        or performing rotation calculations, as discussed in
        :cite:`rowenhorst2015consistent`.
        """
        if np.size(ro) == 0:
            return cls.empty()

        if isinstance(ro, Vector3d):
            ro = ro.data
        else:
            ro = np.atleast_2d(ro)
            if ro.shape[-1] != 3:
                raise ValueError("Final dimension of vector array must be 3.")

        shape = ro.shape[:-1]
        ro = ro.reshape(-1, 3)

        if angles is None:
            norm = Vector3d(ro).norm
            if np.min(norm) < np.finfo(norm.dtype).resolution * 1000:
                warnings.warn(
                    "Max. estimated error is greater than 0.1%. Rodrigues vectors have "
                    "increasing associated errors for small angle rotations. Consider "
                    "creating quaternions in another way."
                )
            angles = 2 * np.arctan(norm)
            angles = angles[:, np.newaxis]
            ax = np.hstack((ro, angles))
        else:
            angles = angles.ravel()[:, np.newaxis]
            ro_axes_angles = np.hstack((ro, angles))
            ax = _conversions.ro2ax(ro_axes_angles)

        if np.rad2deg(np.max(angles)) > 179.999:
            warnings.warn(
                "Highest angle is greater than 179.999 degrees. Rodrigues vectors "
                "cannot parametrize 2-fold rotations. Consider creating quaternions"
                " in another way."
            )

        Q = cls.from_axes_angles(ax[:, :3], ax[:, 3])
        Q = Q.reshape(*shape)
        Q = Q.unit

        return Q

    # TODO: Remove decorator, **kwargs, and use of "convention" in 0.13
    @classmethod
    @deprecated_argument("convention", "0.9", "0.13", "direction")
    def from_euler(
        cls,
        euler: Union[np.ndarray, tuple, list],
        direction: str = "lab2crystal",
        degrees: bool = False,
        **kwargs,
    ) -> Quaternion:
        """Create unit quaternions from Euler angle sets
        :cite:`rowenhorst2015consistent`.

        Parameters
        ----------
        euler
            Euler angles in radians (``degrees=False``) or in degrees
            (``degrees=True``) in the Bunge convention.
        direction
            Direction of the transformation, either ``"lab2crystal"``
            (default) or the inverse, ``"crystal2lab"``. The former is
            the Bunge convention. Passing ``"MTEX"`` equals the latter.
        degrees
            If ``True``, the given angles are assumed to be in degrees.
            Default is ``False``.

        Returns
        -------
        Q
            Unit quaternions.
        """
        direction = direction.lower()
        if direction == "mtex" or (
            "convention" in kwargs and kwargs["convention"] == "mtex"
        ):
            # MTEX' rotations are transformations from the crystal to
            # the lab reference frames. See
            # https://mtex-toolbox.github.io/MTEXvsBungeConvention.html
            # and orix issue #215
            direction = "crystal2lab"

        directions = ["lab2crystal", "crystal2lab"]

        # processing directions
        if direction not in directions:
            raise ValueError(
                f"The chosen direction is not one of the allowed options {directions}"
            )

        if degrees:
            euler = np.deg2rad(euler)

        eu = np.atleast_2d(euler)
        if np.any(np.abs(eu) > 4 * np.pi):
            warnings.warn("Angles are quite high, did you forget to set degrees=True?")

        Q = _conversions.eu2qu(eu)
        Q = cls(Q)

        if direction == "crystal2lab":
            Q = ~Q

        return Q

    @classmethod
    def from_matrix(cls, matrix: Union[np.ndarray, tuple, list]) -> Quaternion:
        """Create unit quaternions from orientation matrices
        :cite:`rowenhorst2015consistent`.

        Parameters
        ----------
        matrix
            Sequence of orientation matrices with the last two
            dimensions of shape ``(3, 3)``.

        Returns
        -------
        Q
            Unit quaternions.

        Examples
        --------
        >>> from orix.quaternion import Quaternion
        >>> Q = Quaternion.from_matrix([np.eye(3), 2 * np.eye(3), np.diag([1, -1, -1])])
        >>> Q
        Quaternion (3,)
        [[1. 0. 0. 0.]
         [1. 0. 0. 0.]
         [0. 1. 0. 0.]]
        """
        # Verify the input can be interpreted as an array of (3, 3) arrays
        om = np.atleast_2d(matrix)
        if om.shape[-2:] != (3, 3):
            raise ValueError("the last two dimensions of 'matrix' must be (3, 3)")

        Q = _conversions.om2qu(om)
        Q = cls(Q)

        return Q

    @classmethod
    def from_scipy_rotation(cls, rotation: SciPyRotation) -> Quaternion:
        """Create unit quaternions from
        :class:`scipy.spatial.transform.Rotation`.

        Parameters
        ----------
        rotation
            SciPy rotations.

        Returns
        -------
        quaternion
            Quaternions.

        Notes
        -----
        The SciPy rotation is inverted to be consistent with the orix
        framework of passive rotations.

        While orix represents quaternions with the scalar as the first
        parameter, SciPy has the scalar as the last parameter.

        Examples
        --------
        >>> from orix.quaternion import Quaternion
        >>> from orix.vector import Vector3d
        >>> from scipy.spatial.transform import Rotation as SciPyRotation

        SciPy and orix represent quaternions differently

        >>> R_scipy = SciPyRotation.from_euler("ZXZ", [90, 0, 0], degrees=True)
        >>> R_scipy.as_quat()
        array([0.        , 0.        , 0.70710678, 0.70710678])
        >>> Q = Quaternion.from_scipy_rotation(R_scipy)
        >>> Q
        Quaternion (1,)
        [[ 0.7071  0.      0.     -0.7071]]
        >>> ~Q
        Quaternion (1,)
        [[ 0.7071 -0.     -0.      0.7071]]

        SciPy and orix rotate vectors differently

        >>> v = [1, 1, 0]
        >>> R_scipy.apply(v)
        array([-1.,  1.,  0.])
        >>> Q * Vector3d(v)
        Vector3d (1,)
        [[ 1. -1.  0.]]
        >>> ~Q * Vector3d(v)
        Vector3d (1,)
        [[-1.  1.  0.]]
        """
        matrix = rotation.inv().as_matrix()
        return cls.from_matrix(matrix=matrix)

    @classmethod
    def from_align_vectors(
        cls,
        other: Union[Vector3d, tuple, list],
        initial: Union[Vector3d, tuple, list],
        weights: Optional[np.ndarray] = None,
        return_rmsd: bool = False,
        return_sensitivity: bool = False,
    ) -> Union[
        Quaternion,
        Tuple[Quaternion, float],
        Tuple[Quaternion, np.ndarray],
        Tuple[Quaternion, float, np.ndarray],
    ]:
        """Estimate a quaternion to optimally align two sets of vectors.

        This method wraps
        :meth:`~scipy.spatial.transform.Rotation.align_vectors`. See
        that method for further explanations of parameters and returns.

        Parameters
        ----------
        other
            Vectors of shape ``(n,)`` in the other reference frame.
        initial
            Vectors of shape ``(n,)`` in the initial reference frame.
        weights
            Relative importance of the different vectors.
        return_rmsd
            Whether to return the (weighted) root mean square distance
            between ``other`` and ``initial`` after alignment. Default
            is ``False``.
        return_sensitivity
            Whether to return the sensitivity matrix. Default is
            ``False``.

        Returns
        -------
        estimated_quaternion
            Best estimate of the quaternion that transforms ``initial``
            to ``other``.
        rmsd
            Returned when ``return_rmsd=True``.
        sensitivity
            Returned when ``return_sensitivity=True``.

        Examples
        --------
        >>> from orix.quaternion import Quaternion
        >>> from orix.vector import Vector3d
        >>> v1 = Vector3d([[1, 0, 0], [0, 1, 0]])
        >>> v2 = Vector3d([[0, -1, 0], [0, 0, 1]])
        >>> Q12 = Quaternion.from_align_vectors(v2, v1)
        >>> Q12 * v1
        Vector3d (2,)
        [[ 0. -1.  0.]
         [ 0.  0.  1.]]
        >>> Q21, dist = Quaternion.from_align_vectors(v1, v2, return_rmsd=True)
        >>> dist
        0.0
        >>> Q21 * v2
        Vector3d (2,)
        [[1. 0. 0.]
         [0. 1. 0.]]
        """
        if not isinstance(other, Vector3d):
            other = Vector3d(other)
        if not isinstance(initial, Vector3d):
            initial = Vector3d(initial)
        v1 = initial.unit.data
        v2 = other.unit.data

        out = SciPyRotation.align_vectors(
            v1, v2, weights=weights, return_sensitivity=return_sensitivity
        )
        out = list(out)
        out[0] = cls.from_scipy_rotation(out[0])

        if not return_rmsd:
            del out[1]

        return out[0] if len(out) == 1 else tuple(out)

    @classmethod
    def triple_cross(cls, q1: Quaternion, q2: Quaternion, q3: Quaternion) -> Quaternion:
        """Pointwise cross product of three quaternions.

        Parameters
        ----------
        q1
            First quaternions.
        q2
            Second quaternions.
        q3
            Third quaternions.

        Returns
        -------
        Q
            Quaternions resulting from the triple cross product.
        """
        Q1a, Q1b, Q1c, Q1d = q1.a, q1.b, q1.c, q1.d
        Q2a, Q2b, Q2c, Q2d = q2.a, q2.b, q2.c, q2.d
        Q3a, Q3b, Q3c, Q3d = q3.a, q3.b, q3.c, q3.d
        # fmt: off
        a = (
            + Q1b * Q2c * Q3d
            - Q1b * Q3c * Q2d
            - Q2b * Q1c * Q3d
            + Q2b * Q3c * Q1d
            + Q3b * Q1c * Q2d
            - Q3b * Q2c * Q1d
        )
        b = (
            + Q1a * Q3c * Q2d
            - Q1a * Q2c * Q3d
            + Q2a * Q1c * Q3d
            - Q2a * Q3c * Q1d
            - Q3a * Q1c * Q2d
            + Q3a * Q2c * Q1d
        )
        c = (
            + Q1a * Q2b * Q3d
            - Q1a * Q3b * Q2d
            - Q2a * Q1b * Q3d
            + Q2a * Q3b * Q1d
            + Q3a * Q1b * Q2d
            - Q3a * Q2b * Q1d
        )
        d = (
            + Q1a * Q3b * Q2c
            - Q1a * Q2b * Q3c
            + Q2a * Q1b * Q3c
            - Q2a * Q3b * Q1c
            - Q3a * Q1b * Q2c
            + Q3a * Q2b * Q1c
        )
        # fmt: on
        Q = cls(np.vstack((a, b, c, d)).T)
        return Q

    @classmethod
    def identity(cls, shape: Union[int, tuple] = (1,)) -> Quaternion:
        """Create identity quaternions.

        Parameters
        ----------
        shape
            Shape of the quaternion instance.

        Returns
        -------
        Q
            Identity quaternions.
        """
        shape = (shape,) if isinstance(shape, int) else shape
        Q = np.zeros(shape + (4,))
        Q[..., 0] = 1
        return cls(Q)

    # ---------------------- All "to_*" methods- --------------------- #

    # TODO: Remove decorator and **kwargs in 0.13
    @deprecated_argument("convention", since="0.9", removal="0.13")
    def to_euler(self, degrees: bool = False, **kwargs) -> np.ndarray:
        r"""Return the unit quaternions as Euler angles in the Bunge
        convention :cite:`rowenhorst2015consistent`.

        Parameters
        ----------
        degrees
            If ``True``, the given angles are returned in degrees.
            Default is ``False``.

        Returns
        -------
        eu
            Array of Euler angles in radians (``degrees=False``) or
            degrees (``degrees=True``), in the ranges
            :math:`\phi_1 \in [0, 2\pi]`, :math:`\Phi \in [0, \pi]`, and
            :math:`\phi_1 \in [0, 2\pi]`.
        """
        eu = _conversions.qu2eu(self.unit.data)
        if degrees:
            eu = np.rad2deg(eu)
        return eu

    def to_matrix(self) -> np.ndarray:
        """Return the unit quaternions as orientation matrices
        :cite:`rowenhorst2015consistent`.

        Returns
        -------
        om
            Array of orientation matrices after normalizing the
            quaternions.

        Examples
        --------
        >>> from orix.quaternion import Quaternion
        >>> Q1 = Quaternion([[1, 0, 0, 0], [2, 0, 0, 0]])
        >>> np.allclose(Q1.to_matrix(), np.eye(3))
        True
        >>> Q2 = Quaternion([[0, 1, 0, 0], [0, 2, 0, 0]])
        >>> np.allclose(Q2.to_matrix(), np.diag([1, -1, -1]))
        True
        """
        om = _conversions.qu2om(self.unit.data)
        return om

    def to_axes_angles(self) -> AxAngle:
        r"""Return the unit quaternions as axis-angle vectors
        :cite:`rowenhorst2015consistent`.

        Returns
        -------
        ax
            Axis-angle vectors with magnitude :math:`\theta` equal to
            the angle of rotation.

        See Also
        --------
        to_homochoric, to_rodrigues

        Examples
        --------
        A 3-fold rotation around the [111] axis

        >>> from orix.quaternion import Quaternion
        >>> Q = Quaternion([0.5, 0.5, 0.5, 0.5])
        >>> ax = Q.to_axes_angles()
        >>> ax
        AxAngle (1,)
        [[1.2092 1.2092 1.2092]]
        >>> np.rad2deg(ax.angle)
        array([120.])
        """
        axes, angles = _conversions.qu2ax(self.unit.data)
        ax = AxAngle(axes * angles)
        return ax

    def to_rodrigues(self, frank: bool = False) -> Union[Rodrigues, np.ndarray]:
        r"""Return the unit quaternions as Rodrigues or Rodrigues-Frank
         vectors :cite:`rowenhorst2015consistent`.

        Parameters
        ----------
        frank
            Whether to return Rodrigues vectors scaled by
            :math:`\tan(\theta/2)`, where :math:`\theta` is the angle of
            rotation, or Rodrigues-Frank vectors scaled by
            :math:`\omega = 2\arctan(|\rho|)` in an array.

        Returns
        -------
        ro
            Vectors :math:`\hat{\mathbf{n}}` parallel to the axis of
            rotation if ``frank=False`` or an array of four-component
            vectors if ``frank=True``.

        See Also
        --------
        to_axes_angles, to_homochoric

        Examples
        --------
        A 3-fold rotation around the [111] axis

        >>> from orix.quaternion import Quaternion
        >>> Q = Quaternion.from_axes_angles([1, 1, 1], 120, degrees=True)
        >>> ro1 = Q.to_rodrigues()
        >>> ro1
        Rodrigues (1,)
        [[1. 1. 1.]]
        >>> ro1.norm
        array([1.73205081])
        >>> ro2 = Q.to_rodrigues(frank=True)
        >>> ro2
        array([[0.57735027, 0.57735027, 0.57735027, 1.73205081]])
        >>> np.linalg.norm(ro2[:, :3])
        1.0

        A 45:math:`\degree` rotation around the [111] axis

        >>> Q2 = Quaternion.from_axes_angles([1, 1, 1], 45, degrees=True)
        >>> ro3 = Q2.to_rodrigues()
        >>> ro3
        Rodrigues (1,)
        [[0.2391 0.2391 0.2391]]

        Notes
        -----
        Rodrigues vectors, originally proposed by O. Rodrigues, are
        often used for plotting orientations as they create isomorphic
        (though not volume-preserving) plots and form fundamental zones
        with rectilinear boundaries. These features are
        well-demonstrated in :cite:`frank1988orientation`. See
        :cite:`rowenhorst2015consistent` for examples of usage of
        Rodrigues-Frank vectors.
        """
        Q = self.unit
        if not frank:
            ro = Q.axis * np.tan(self.angle / 2)
            ro = Rodrigues(ro)
        else:
            axes, angles = _conversions.qu2ax(Q.data)
            axes_angles = np.concatenate((axes, angles), axis=-1)
            ro = _conversions.ax2ro(axes_angles)
        return ro

    def to_homochoric(self) -> Homochoric:
        r"""Return the unit quaternions as homochoric vectors
        :cite:`rowenhorst2015consistent`.

        Returns
        -------
        ho
            Homochoric vectors parallel to the axes of rotation with
            lengths equal to
            :math:`\left[\frac{3}{4}\cdot(\theta - \sin(\theta))\right]^{1/3}`,
            where :math:`\theta` is the angle of rotation.

        See Also
        --------
        to_axes_angles, from_rodrigues

        Examples
        --------
        A 3-fold rotation about the [111] axis

        >>> from orix.quaternion import Quaternion
        >>> Q = Quaternion.from_axes_angles([1, 1, 1], 120, degrees=True)
        >>> ho = Q.to_homochoric()
        >>> ho
        Homochoric (1,)
        [[0.5618 0.5618 0.5618]]

        Notes
        -----
        Homochoric vectors are often used for plotting orientations as
        they create an isomorphic (though not angle-preserving) mapping
        from the non-euclidean orientation space into Cartesian
        coordinates. Additionally, unlike Rodrigues vectors, all
        rotations map into a finite space, bounded by a sphere of radius
        :math:`\pi`.
        """
        ho = _conversions.qu2ho(self.unit.data)
        ho = Homochoric(ho)
        return ho

    # --------------------- Other public methods --------------------- #

    def dot(self, other: Quaternion) -> np.ndarray:
        """Return the dot products of the quaternions and the other
        quaternions.

        Parameters
        ----------
        other
            Other quaternions.

        Returns
        -------
        dot_products
            Dot products.

        See Also
        --------
        Rotation.dot
        Orientation.dot

        Examples
        --------
        >>> from orix.quaternion import Quaternion
        >>> Q1 = Quaternion([[1, 0, 0, 0], [0.9239, 0, 0, 0.3827]])
        >>> Q2 = Quaternion([[0.9239, 0, 0, 0.3827], [0.7071, 0, 0, 0.7071]])
        >>> Q1.dot(Q2)
        array([0.9239    , 0.92389686])
        """
        return np.sum(self.data * other.data, axis=-1)

    def dot_outer(self, other: Quaternion) -> np.ndarray:
        """Return the dot products of all quaternions to all the other
        quaternions.

        Parameters
        ----------
        other
            Other quaternions.

        Returns
        -------
        dot_products
            Dot products.

        See Also
        --------
        Rotation.dot_outer
        Orientation.dot_outer

        Examples
        --------
        >>> from orix.quaternion import Quaternion
        >>> Q1 = Quaternion([[1, 0, 0, 0], [0.9239, 0, 0, 0.3827]])
        >>> Q2 = Quaternion([[0.9239, 0, 0, 0.3827], [0.7071, 0, 0, 0.7071]])
        >>> Q1.dot_outer(Q2)
        array([[0.9239    , 0.7071    ],
               [1.0000505 , 0.92389686]])
        """
        dots = np.tensordot(self.data, other.data, axes=(-1, -1))
        return dots

    def mean(self) -> Quaternion:
        """Return the mean quaternion with unitary weights.

        Returns
        -------
        quat_mean
            Mean quaternion.

        Notes
        -----
        The method used here corresponds to Equation (13) in
        https://arc.aiaa.org/doi/pdf/10.2514/1.28949.
        """
        Q = self.flatten().data.T
        QQ = Q.dot(Q.T)
        w, v = np.linalg.eig(QQ)
        w_max = np.argmax(w)
        return self.__class__(v[:, w_max])

    def outer(
        self,
        other: Union[Quaternion, Vector3d],
        lazy: bool = False,
        chunk_size: int = 20,
        progressbar: bool = True,
    ) -> Union[Quaternion, Vector3d]:
        """Return the outer products of the quaternions and the other
        quaternions or vectors.

        Parameters
        ----------
        other
            Another orientation or vector.
        lazy
            Whether to computer this computation using Dask. This option
            can be used to reduce memory usage when working with large
            arrays. Default is ``False``.
        chunk_size
            When using ``lazy`` computation, ``chunk_size`` represents
            the number of objects per axis for each input to include in
            each iteration of the computation. Default is 20.
        progressbar
            Whether to show a progressbar during computation if
            ``lazy=True``. Default is ``True``.

        Returns
        -------
        out
            Outer products.

        Raises
        ------
        NotImplementedError
            If ``other`` is not a quaternion, 3D vector, or a Miller
            index.
        """
        if isinstance(other, Quaternion):
            if lazy:
                darr = self._outer_dask(other, chunk_size=chunk_size)
                arr = np.empty(darr.shape)
                if progressbar:
                    with ProgressBar():
                        da.store(darr, arr)
                else:
                    da.store(darr, arr)
            else:
                Q1 = quaternion.from_float_array(self.data)
                Q2 = quaternion.from_float_array(other.data)
                # np.outer works with flattened array
                Q = np.outer(Q1, Q2).reshape(Q1.shape + Q2.shape)
                arr = quaternion.as_float_array(Q)
            return other.__class__(arr)
        elif isinstance(other, Vector3d):
            if lazy:
                darr = self._outer_dask(other, chunk_size=chunk_size)
                arr = np.empty(darr.shape)
                if progressbar:
                    with ProgressBar():
                        da.store(darr, arr)
                else:
                    da.store(darr, arr)
            else:
                Q = quaternion.from_float_array(self.data)
                arr = quaternion.rotate_vectors(Q, other.data)
            if isinstance(other, Miller):
                m = other.__class__(xyz=arr, phase=other.phase)
                m.coordinate_format = other.coordinate_format
                return m
            else:
                return other.__class__(arr)
        else:
            raise NotImplementedError(
                "This operation is currently not avaliable in orix, please use outer "
                "with `other` of type `Quaternion` or `Vector3d`"
            )

    def inv(self) -> Quaternion:
        r"""Return the inverse quaternions
        :math:`Q^{-1} = a - bi - cj - dk`.
        """
        return self.__invert__()

    # -------------------- Other private methods --------------------- #

    def _outer_dask(
        self, other: Union[Quaternion, Vector3d], chunk_size: int = 20
    ) -> da.Array:
        """Compute the product of every quaternion in this instance to
        every quaternion or vector in another instance, returned as a
        Dask array.

        For quaternion-quaternion multiplication, this is known as the
        Hamilton product.

        Parameters
        ----------
        other
            Another orientation or vector.
        chunk_size
            Number of objects per axis for each input to include in each
            iteration of the computation. Default is 20.

        Returns
        -------
        out

        Raises
        ------
        TypeError
            If ``other`` is not a quaternion or a vector.

        Notes
        -----
        For quaternion-quaternion multiplication, to create a new
        quaternion from the returned array ``out``, do
        ``q = Quaternion(out.compute())``. Likewise for
        quaternion-vector multiplication, to create a new vector from
        the returned array do ``v = Vector3d(out.compute())``.
        """
        if not isinstance(other, (Quaternion, Vector3d)):
            raise TypeError("Other must be Quaternion or Vector3d.")

        ndim1 = len(self.shape)
        ndim2 = len(other.shape)

        # Set chunk sizes
        chunks1 = (chunk_size,) * ndim1 + (-1,)
        chunks2 = (chunk_size,) * ndim2 + (-1,)

        # Dask has no dask.multiply.outer(), use dask.array.einsum
        # Summation subscripts
        str1 = "abcdefghijklm"[:ndim1]  # Max. object dimension of 13
        str2 = "nopqrstuvwxyz"[:ndim2]
        sum_over = f"...{str1},{str2}...->{str1 + str2}"

        # Get quaternion parameters as dask arrays to be computed later
        Q1 = da.from_array(self.data, chunks=chunks1)
        a1, b1, c1, d1 = Q1[..., 0], Q1[..., 1], Q1[..., 2], Q1[..., 3]

        # We silence dask's einsum performance warnings for "small"
        # chunk sizes, since using the chunk sizes suggested floods
        # memory
        warnings.filterwarnings("ignore", category=da.PerformanceWarning)

        if isinstance(other, Quaternion):
            Q2 = da.from_array(other.data, chunks=chunks2)
            a2, b2, c2, d2 = Q2[..., 0], Q2[..., 1], Q2[..., 2], Q2[..., 3]
            # fmt: off
            a = (
                + da.einsum(sum_over, a1, a2)
                - da.einsum(sum_over, b1, b2)
                - da.einsum(sum_over, c1, c2)
                - da.einsum(sum_over, d1, d2)
            )
            b = (
                + da.einsum(sum_over, b1, a2)
                + da.einsum(sum_over, a1, b2)
                - da.einsum(sum_over, d1, c2)
                + da.einsum(sum_over, c1, d2)
            )
            c = (
                + da.einsum(sum_over, c1, a2)
                + da.einsum(sum_over, d1, b2)
                + da.einsum(sum_over, a1, c2)
                - da.einsum(sum_over, b1, d2)
            )
            d = (
                + da.einsum(sum_over, d1, a2)
                - da.einsum(sum_over, c1, b2)
                + da.einsum(sum_over, b1, c2)
                + da.einsum(sum_over, a1, d2)
            )
            # fmt: on
            out = da.stack([a, b, c, d], axis=-1)
        else:  # Vector3d
            v2 = da.from_array(other.data, chunks=chunks2)
            x2, y2, z2 = v2[..., 0], v2[..., 1], v2[..., 2]
            # fmt: off
            x = (
                + da.einsum(sum_over, a1**2 + b1**2 - c1**2 - d1**2, x2)
                + da.einsum(sum_over, a1 * c1 + b1 * d1, z2) * 2
                + da.einsum(sum_over, b1 * c1 - a1 * d1, y2) * 2
            )
            y = (
                + da.einsum(sum_over, a1**2 - b1**2 + c1**2 - d1**2, y2)
                + da.einsum(sum_over, a1 * d1 + b1 * c1, x2) * 2
                + da.einsum(sum_over, c1 * d1 - a1 * b1, z2) * 2
            )
            z = (
                + da.einsum(sum_over, a1**2 - b1**2 - c1**2 + d1**2, z2)
                + da.einsum(sum_over, a1 * b1 + c1 * d1, y2) * 2
                + da.einsum(sum_over, b1 * d1 - a1 * c1, x2) * 2
            )
            # fmt: on
            out = da.stack([x, y, z], axis=-1)

        new_chunks = tuple(chunks1[:-1]) + tuple(chunks2[:-1]) + (-1,)

        return out.rechunk(new_chunks)
